/*
   init.c -- PIKT initialization routines

   $Header: /home/cvs/ucgsb/pikt/src/lib/init.c,v 1.8 1999/10/13 15:05:33 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "common.h"

FILE *confin;

int ismaster;

char *homdir = (char *) NULL;	/* be explicit about initialization, just
				   to be sure (also for all that follow)
				 */
char *bindir = (char *) NULL;
char *etcdir = (char *) NULL;
char *libdir = (char *) NULL;
char *vardir = (char *) NULL;
char *altdir = (char *) NULL;
char *prgdir = (char *) NULL;
char *objdir = (char *) NULL;
char *cfgdir = (char *) NULL;
char *difdir = (char *) NULL;
char *stgdir = (char *) NULL;
char *hstdir = (char *) NULL;
char *logdir = (char *) NULL;
char *PIKTconf = (char *) NULL;
char *pikt = (char *) NULL;
char *piktd = (char *) NULL;
char *piktc = (char *) NULL;
char *piktc_svc = (char *) NULL;
char *piktdconf = (char *) NULL;
char *piktdconftmp = (char *) NULL;
char *piktdlock = (char *) NULL;
char *piktdlog = (char *) NULL;
char *piktdalt = (char *) NULL;
char *piktclock = (char *) NULL;
char *piktclog = (char *) NULL;
char *piktc_svclock = (char *) NULL;
char *piktc_svclog = (char *) NULL;

unsigned long confparsrv;
char **confparstrp;
int *confparintp;

int isopt(char **, char **, int);
void gethome(char **);
void initpaths(char *);
void *initmalloc(size_t, char *);

/* perform various initializations
 */
void
init (int argc, char **argv, void (*usage)(void))
{
	char *cp;
	char *fqhn;

	istest = isopt(argv+1, argv+argc-1, 'T');
	isdebug = isopt(argv+1, argv+argc-1, 'G');
	if((ishelp = isopt(argv+1, argv+argc-1, 'h'))) {
		(*usage)();
		exit(1);
	}

	gethostname(hostname, MAXIDENT);
	cp = strchr(hostname, '.');
	if (cp) {
		fqhn = strdup(hostname);
		if (! istest)
			*cp = '\0';
		strcpy(domainname, cp+1);
	} else {
		fqhn = (char *)NULL;
		/* what we really want here is not the nis domain,
		   but rather the internet domainname
		 */
		#ifdef SOLARIS
		sysinfo(SI_SRPC_DOMAIN, domainname, MAXIDENT);
		#else
		getdomainname(domainname, MAXIDENT);
		#endif
	}

	gethome(argv);		/* determine PIKT home dir */

	if (istest) {
		master = strdup(hostname);
		domain = strdup("domain");
		accessmode = strdup("local");
	} else {
		getconf(homdir);	/* get config stuff from PIKT.conf */
	}

	if (bailout)
		return;

	initpaths(homdir);	/* initialize dir/file paths */

	if (bailout)
		return;

	/* from here on, it is safe to call either logreturn() or logexit()
	   since we have the paths to the log files
	 */

	if (fqhn) {
		if (!strcmp(fqhn, master))
			ismaster = TRUE;
		free(fqhn);
	} else {
		if (!strncmp(hostname, master, (size_t)strlen(hostname)))
			ismaster = TRUE;
	}

	sig_init();		/* initialize signal handlers */

	return;
}

/* look for a command-line (minus) switch, returning TRUE if found, else
   FALSE; this is for simpler command-line parsing; for more involved parsing,
   we use cmd.[lyc] and cmdparse()
 */
int
isopt (char **avcurr, char **avlast, int opt)
{
	while (avcurr <= avlast) {
		if (**avcurr == '-') {
			if (strchr(*avcurr, opt))
				return TRUE;
		}
		avcurr++;
	}
	return FALSE;
}

/* determine the PIKT home directory, either directly from argv (when
   the program is invoked with a full path), or by searching through the
   environment PATH (when the program is invoked with no path)
 */
void
gethome (char **progpath)
{
	char *pp;
	static char *envpath, *path, *prog, *pathprog;
	int proglen, homdirlen;
	struct stat fbuf;
	char dircur[MAXPATH+1];
	int found;

	/* note that for error logging, we can't invoke logreturn() or
	   logexit(), because we don't have the path to the log file yet!
	 */

	homdir = prog = strdup(*progpath);
	proglen = strlen(prog);
	if ((pp = strrchr(homdir, '/'))) {
		*pp = '\0';
		if (!strcmp(homdir, ".")) {	/* prog invoked as ./<prog> */
			if (getcwd(dircur, MAXPATH) == NULL) {
				fprintf(stderr, "in gethome(), getcwd() failure, cannot determine current working directory\n");
				fflush(stderr);
				exit(1);
			}
			free(homdir);
			homdir = strdup(dircur);
		}
	} else {	/* there is no explicit path, so we have to get it from
	           the PATH environment variable
	         */
		if (! (pp = getenv("PATH"))) {
			fprintf(stderr, "in gethome(), cannot determine environment PATH\n");
			fflush(stderr);
			exit(1);
		} else
			envpath = strdup(pp);
		/* traverse the PATH, until we find PIKT home */
		if (strtok(envpath, ": ") == NULL) {
			fprintf(stderr, "in gethome(), cannot tokenize environment PATH\n");
			fflush(stderr);
			exit(1);
		} else
			path = envpath;
		do {
			pathprog = (char *) initmalloc((strlen(path)+2+proglen) * sizeof(char), "gethome");
			sprintf(pathprog, "%s/%s", path, prog);
			found = ! lstat(pathprog, &fbuf);
			free(pathprog);
			if (found) {
				free(prog);
				homdir = strdup(path);
				break;
			}
		} while ((path = strtok(NULL, ": ")));
		if (! path) {
			fprintf(stderr, "in gethome(), %s not in environment PATH\n", prog);
			fflush(stderr);
			exit(1);
		}
		free(envpath);
	}
	homdirlen = strlen(homdir);
	if ((homdirlen >= 4) && ! strcmp(homdir+homdirlen-4, "/bin"))
		*(homdir+homdirlen-4) = '\0';
	else {
		free(homdir);
		homdir = strdup("");
	}
	return;
}

/* get PIKT.conf file settings
 */
void
getconf (char *hom)
{
	const char *testbranch;

	if (istest)
		testbranch = "/src/test";
	else
		testbranch = "";
	PIKTconf = (char *) initmalloc((strlen(hom)+strlen(testbranch)+15) * sizeof(char), "getconf");
	sprintf(PIKTconf, "%s%s/etc/PIKT.conf", hom, testbranch);
	if ((confin = fopen(PIKTconf, "r")) == NULL) {
		free(PIKTconf);
		PIKTconf = strdup("/etc/PIKT.conf");
		if ((confin = fopen(PIKTconf, "r")) == NULL) {
			fprintf(stderr, "in getconf(), fopen() error for (possibly missing) PIKT.conf file\n");
			fflush(stderr);
			exit(1);
		}
	}
	lineno = 1;
	confparse();
	if (fclose(confin)) {
		fprintf(stderr, "in getconf(), fclose() error for %s\n", PIKTconf);
		fflush(stderr);
		/* bailout = TRUE; */
	}
	if (bailout)
		return;
	/* now check for the presence of certain essential elements */
	if (!master) {
		fprintf(stderr, "in PIKT.conf, failure to specify master\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (master && !strchr(master, '.')) {
		fprintf(stderr, "in PIKT.conf, master is not fully qualified network name\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (!domain) {
		fprintf(stderr, "in PIKT.conf, failure to specify domain\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (!accessmode) {
		fprintf(stderr, "in PIKT.conf, failure to specify access\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (isauth_by_master_address && !master_address) {
		fprintf(stderr, "in PIKT.conf, failure to specify master address\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (!isauth_by_master && !isauth_by_master_address) {
		fprintf(stderr, "in PIKT.conf, authenticate_by_master and/or authenticate_by_master_address\nmust be YES (TRUE or ON)\n");
		fflush(stderr);
		bailout = TRUE;
	}
	return;	
}

/* from PIKT.conf, set the string parameter
 */
void
setstr (char *name, char **parameter, char *setting)
{
	*parameter = strdup(setting);
	return;
}

/* from PIKT.conf, set/unset the specified option flag, in the case of
   boolean options, else set the numerical parameter
 */
void
setint (char *name, int *parameter, int setting)
{
	*parameter = setting;
	return;
}

/* from PIKT.conf, set/unset the specified bit in services, thereby
   enabling/disabling the specified piktc_svc service
 */
void
setsrv (char *name, unsigned long service, int setting)
{

	if (setting == TRUE)
		services |= service;
	else if (setting == FALSE)
		services &= ~service;
	else {
		fprintf(stderr, "in setsrv(), for service %s, invalid setting %d\n", name, setting);
		fflush(stderr);
	}
	return;
}

/* initialize dir/file paths
 */
void
initpaths (char *hom)
{
	const char *testbranch;

	if (istest)
		testbranch = "/src/test";
	else
		testbranch = "";

	/* there are lots of redundant calls to strlen() here; we could
	   save strlen() return values to local vars, but in the interest
	   of simplicity we choose not to
	 */

	if (! bindir) {
		bindir = (char *) initmalloc((strlen(hom)+5) * sizeof(char), "initpaths");
		sprintf(bindir, "%s/bin", hom);	/* we don't insert
						   testbranch into bindir
						 */
	}

	if (! etcdir ) {
		etcdir = (char *) initmalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), "initpaths");
		sprintf(etcdir, "%s%s/etc", hom, testbranch);
	}

	if (! libdir) {
		libdir = (char *) initmalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), "initpaths");
		sprintf(libdir, "%s%s/lib", hom, testbranch);
	}

	if (! vardir ) {
		vardir = (char *) initmalloc((strlen(hom)+strlen(testbranch)+5) * sizeof(char), "initpaths");
		sprintf(vardir, "%s%s/var", hom, testbranch);
	}

	if (! altdir) {
		altdir = (char *) initmalloc((strlen(libdir)+8) * sizeof(char), "initpaths");
		sprintf(altdir, "%s/alerts", libdir);
	}

	if (! prgdir) {
		prgdir = (char *) initmalloc((strlen(libdir)+10) * sizeof(char), "initpaths");
		sprintf(prgdir, "%s/programs", libdir);
	}

	if (! objdir ) {
		objdir = (char *) initmalloc((strlen(libdir)+9) * sizeof(char), "initpaths");
		sprintf(objdir, "%s/objects", libdir);
	}

	if (! cfgdir ) {
		cfgdir = (char *) initmalloc((strlen(libdir)+9) * sizeof(char), "initpaths");
		sprintf(cfgdir, "%s/configs", libdir);
	}

	/* note: for now, one cannot override difdir or stgdir in PIKT.conf--
	   they are always assumed to be direct subdirectories off cfgdir
	 */

	if (! difdir) {
		difdir = (char *) initmalloc((strlen(cfgdir)+9) * sizeof(char), "initpaths");
		sprintf(difdir, "%s/diffing", cfgdir);
	}

	if (! stgdir) {
		stgdir = (char *) initmalloc((strlen(cfgdir)+9) * sizeof(char), "initpaths");
		sprintf(stgdir, "%s/staging", cfgdir);
	}

	if (! hstdir) {
		hstdir = (char *) initmalloc((strlen(vardir)+11) * sizeof(char), "initpaths");
		sprintf(hstdir, "%s/histories", vardir);
	}

	if (! logdir) {
		logdir = (char *) initmalloc((strlen(vardir)+5) * sizeof(char), "initpaths");
		sprintf(logdir, "%s/log", vardir);
	}

	/* already determined in getconf()
	PIKTconf = (char *) initmalloc((strlen(etcdir)+11) * sizeof(char), "initpaths");
	sprintf(PIKTconf, "%s/PIKT.conf", etcdir);
	 */

	pikt = (char *) initmalloc((strlen(bindir)+6) * sizeof(char), "initpaths");
	sprintf(pikt, "%s/pikt", bindir);

	piktd = (char *) initmalloc((strlen(bindir)+7) * sizeof(char), "initpaths");
	sprintf(piktd, "%s/piktd", bindir);

	piktc = (char *) initmalloc((strlen(bindir)+7) * sizeof(char), "initpaths");
	sprintf(piktc, "%s/piktc", bindir);

	piktc_svc = (char *) initmalloc((strlen(bindir)+11) * sizeof(char), "initpaths");
	sprintf(piktc_svc, "%s/piktc_svc", bindir);

	piktdconf = (char *) initmalloc((strlen(etcdir)+12) * sizeof(char), "initpaths");
	sprintf(piktdconf, "%s/piktd.conf", etcdir);

	piktdconftmp = (char *) initmalloc((strlen(piktdconf)+5) * sizeof(char), "initpaths");
	sprintf(piktdconftmp, "%s.tmp", piktdconf);

	piktdlock = (char *) initmalloc((strlen(etcdir)+12) * sizeof(char), "initpaths");
	sprintf(piktdlock, "%s/piktd.lock", etcdir);

	piktdlog = (char *) initmalloc((strlen(logdir)+11) * sizeof(char), "initpaths");
	sprintf(piktdlog, "%s/piktd.log", logdir);

	piktdalt = (char *) initmalloc((strlen(etcdir)+8) * sizeof(char), "initpaths");
	sprintf(piktdalt, "%s/alerts", etcdir);

	piktclock = (char *) initmalloc((strlen(etcdir)+12) * sizeof(char), "initpaths");
	sprintf(piktclock, "%s/piktc.lock", etcdir);

	piktclog = (char *) initmalloc((strlen(logdir)+11) * sizeof(char), "initpaths");
	sprintf(piktclog, "%s/piktc.log", logdir);

	piktc_svclock = (char *) initmalloc((strlen(etcdir)+16) * sizeof(char), "initpaths");
	sprintf(piktc_svclock, "%s/piktc_svc.lock", etcdir);

	piktc_svclog = (char *) initmalloc((strlen(logdir)+15) * sizeof(char), "initpaths");
	sprintf(piktc_svclog, "%s/piktc_svc.log", logdir);

	/* now check for the presence of certain essential filepaths */
	if (!isfile(homdir)) {
		fprintf(stderr, "in getconf(), homdir not found\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (!isfile(bindir)) {
		fprintf(stderr, "in getconf(), bindir not found\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (!isfile(etcdir)) {
		fprintf(stderr, "in getconf(), etcdir not found\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (!isfile(libdir)) {
		fprintf(stderr, "in getconf(), libdir not found\n");
		fflush(stderr);
		bailout = TRUE;
	}
	if (!isfile(vardir)) {
		fprintf(stderr, "in getconf(), vardir not found\n");
		fflush(stderr);
		bailout = TRUE;
	}
	return;
}

/* do the first and second given host names match?
   'vienna' matches 'vienna'
   'vienna' matches 'vienna.uppity.edu'
   'vienna.uppity.edu' matches 'vienna'
   'acme.com' matches 'acme.com.uk' and vice-versa--practically speaking,
   is this a problem?  we could check for this weird case, but at the cost
   of slowing down all match checks
 */
int
hostmatch (char *hn1, char *hn2)
{
	register int i;
	register char c1, c2;

	i = 0;
	do {
		c1 = *(hn1+i);
		c2 = *(hn2+i);
		if (!c1 || !c2)
			break;
		if (c1 != c2)
			return FALSE;
		i++;
	} while (TRUE);
	if (!c1) {
		if (!c2 || (c2 == '.'))
			return TRUE;
	} else if (!c2) {
		if (!c1 || (c1 == '.'))
			return TRUE;
	}
	return FALSE;
}

/* malloc() the specified amount of memory; if unsuccessful, presumably
   because of virtual memory exhaustion, output a warning to stderr and
   keep trying; if successful, return a generic pointer to the newly
   allocated memory; we can't use mymalloc(), because it calls logreturn(),
   and we don't have the path to the logdir yet
 */
void *
initmalloc (size_t size, char *func)
{
	void *ptr;
	static char msg[64];	/* should be plenty */

	while (! (ptr = malloc(size))) {
		sprintf(msg, "in %s(), malloc() failure", func);
		fprintf(stderr, msg);
		fflush(stderr);
		sleep(MALLOC_TIME_TO_SLEEP);
	}
	return ptr;
}

/* return TRUE if file/dir exists, FALSE if not
 */
int
isfile (char *file)
{
	struct stat fbuf;

	if (stat(file, &fbuf)) {
		if (logdir) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in isfile(), stat() error for %s", file);
		} else {
			fprintf(stderr, "in isfile(), stat() error for %s\n", file);
			fflush(stderr);
		}
		return FALSE;
	} else
		return TRUE;
}
