%{

/* 
   tag.l -- PIKT piktc tag lexer:  deals with machine conditionals and the
            tagging of config file lines; tagging does not apply to
            systems.cfg!

   $Header: /home/cvs/ucgsb/pikt/src/piktc/tag.l,v 1.24 1999/10/17 05:58:50 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "../lib/common.h"
#include "piktc.h"
#include "tag.tab.h"

#define RETURN(token) taglval.sval = strdup(yytext); return token

char *cp;

typedef struct {
	char *file;		/* we should maybe garbage collect these
				   after preprocess()'ing, but this won't
				   gain us much, so why bother
				 */
	FILE *yyin;
	int fd[2];
	YY_BUFFER_STATE bs;
	pid_t pid;
	int domcr;
	int lineno;
} incfil;

incfil incfiles[MAXIFS+1];

char *incfile = (char *) NULL;

void beginincfile(char *);
void endincfile(void);
void pushincfile(char *, FILE *, int *, YY_BUFFER_STATE, pid_t, int, int);
void popincfile(char **, FILE **, int *, YY_BUFFER_STATE *, pid_t *, int *, int *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%x REM COND INC

%%

<*>^#{W}*(fix|freeze){W}*(\177)?\n	{ lineno++;
				  if (!prescan)
					domacros = FALSE;
				}
<*>^#{W}*(unfix|unfreeze){W}*(\177)?\n? { lineno++;
				  if (!prescan)
					domacros = TRUE;
				}

<*>^#{W}*include{W}+\<[^\>]+\>{W}*(\177)?\n? { lineno++;
				  *strrchr(yytext, '>') = '\0';
				  cp = strchr(yytext, '<') + 1;
				  /* linux chokes on this
				  FREE(incfile);
				   */
				  incfile = (char *) mymalloc((strlen(cfgdir)+strlen(cp)+2) * sizeof(char), "taglex");
				  sprintf(incfile, "%s/%s", cfgdir, cp);
				  beginincfile(incfile);
				}

<*>^#{W}*include{W}+\"[^\"]+\"{W}*(\177)?\n? { lineno++;
				  *strrchr(yytext, '\"') = '\0';
				  cp = strchr(yytext, '\"') + 1;
				  /* linux chokes on this
				  FREE(incfile);
				   */
				  incfile = strdup(cp);
				  beginincfile(incfile);
				}

<*>^#{W}*if{W}+			{ BEGIN(COND);
				  return _IF;
				}
<*>^#{W}*(elif|elsif|elseif){W}+	{ BEGIN(COND);
				  return _ELIF;
				}
<*>^#{W}*else{W}+			{ BEGIN(COND);
				  return _ELSE;
				}
<*>^#{W}*else(\177)?\n		{ lineno++;
				  return _ELSE;
				}
<*>^#{W}*(endif|fi){W}+		{ BEGIN(COND);
				  return _ENDIF;
				}
<*>^#{W}*(endif|fi)(\177)?\n?	{ lineno++;
				  return _ENDIF;
				}
<*>^#.?				{ BEGIN(REM);
				  yyless(0);
				}

<*>^ERR.*\n?			{ BEGIN(REM);
				  yyless(3);
				}

<REM>.*\n?			{ if ((cp = strchr(yytext,'\n')))
					*cp = '\0';
				  tagerror("syntax error");
				  lineno++;
				}

<COND,INC>[^!&|() \t\n\177]+	{ RETURN(_IDENT); }
<COND,INC>"!"|"&"|"|"|"("|")"	{ return yytext[0]; }
<COND,INC>{W}+			{ ; }
<COND,INC>(\177)?\n		{ BEGIN(INITIAL);
				  lineno++;
				}

<*>^{W}*\n?			{ lineno++;
				  if ((yyleng >1) && ((cfgcurr == PRGCFG) || (cfgcurr == FILCFG))) {
					taglval.sval = strdup("\t\n");
					return _LINE;
				  }
				}

<*>\177\n?			{ lineno++; }

<*>^#{W}*(def|define|undef|undefine|ifdef|ifndef|elifdef|elsifdef|elseifdef|elifndef|elsifndef|elseifndef|elsedef|endifdef|fidef)[^\n\177]*(\177)?\n? {
				  lineno++;
				  /* dispose of any '\177' */
				  if ((cp = strchr(yytext, '\177'))) {
					*cp = *(cp+1);
					*(cp+1) = '\0';
				  }
				  RETURN(_DEF);
				}

<*>^[^#\n\177][^\n\177]*(\177)?\n? {
				  lineno++;
				  /* dispose of any '\177' */
				  if ((cp = strchr(yytext, '\177'))) {
					*cp = *(cp+1);
					*(cp+1) = '\0';
				  }
				  RETURN(_LINE);
				}

%%

/* tag error handler
 */
void
tagerror (const char *message)
{

	fprintf(stderr, "%s, at or near line %d: %s (in tagparse()): %s\n", cfgfilecurr, lineno, message, yytext);
	fflush(stderr);
	bailout = TRUE;
	if (incnum == 0)
		BEGIN(INITIAL);
	return;
}

/* tag end-of-file handler
 */
int
tagwrap (void)
{

	if (incnum > 0) {
		endincfile();
		return 0;
	} else
		return 1;
}

/* push the current input buffer onto the file stack; create and switch to
   a new input buffer; open the new include file and start feeding its
   contents to the new input buffer; if stripping comments from the new
   include file, spawn a child process to pass the include file
   contents to the strip lexer
 */
void
beginincfile (char *file)
{
	struct stat fbuf;

	if (stat(file, &fbuf) != 0) {
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginincfile(), stat() error for %s", file);
	}
	pushincfile(infile, tagin, fd, YY_CURRENT_BUFFER, pid, domacros, lineno);
	/* linux possibly chokes on this
	FREE(cfgfileprev);
	 */
	cfgfileprev = cfgfilecurr;
	cfgfilecurr = infile = strdup(file);
	if (!prescan)
		domacros = TRUE;
	incnum++;
	if (incnum > MAXIFS)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in beginincfile(), maximum number (%d) of include files exceeded", MAXIFS);
	lineno = 1;
	BEGIN(INC);
	if (dostrip) {
		if (pipe(fd) < 0)
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginincfile(), pipe() error");
		/* open the ipc channels */
		ipcopen(incnum);
		switch (pid = fork()) {
			case -1:	/* error */
				logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in beginincfile(), fork() error");
				break;
			/* pipe from child to parent */
			case 0:		/* child */
				close(fd[0]);
				stripin = fopen(infile, "r");
				stripout = fdopen(fd[1], "w");
				wait_parent(incnum);
				tell_parent(incnum);
				striplex();
				fclose(stripin);
				fclose(stripout);
				close(fd[1]);
				tell_parent(incnum);
				wait_parent(incnum);
				_exit(0);
				break;
			default:	/* parent */
				close(fd[1]);
				tagin = fdopen(fd[0], "r");
				tell_child(incnum);
				wait_child(incnum);
				break;
		}
	} else {
		pid = 0;
		fd[0] = -1;
		fd[1] = -1;
		tagin = fopen(infile, "r");
	}
	yy_switch_to_buffer(yy_create_buffer(tagin, MAXBUF));
	return;
}

/* at the end of the current include file, close the file, delete its
   input buffer, pop the previous include file off the file stack, and
   switch to the previous input buffer
 */
void
endincfile (void)
{
	YY_BUFFER_STATE bs;
	int status;

	fclose(tagin);
	if (dostrip) {
		close(fd[0]);
		wait_child(incnum);
		tell_child(incnum);
		waitpid(pid, &status, 0);
		/* close the ipc channels */
		ipcclose(incnum);
	}
	incnum--;
	if (incnum < 0)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in endincfile(), incnum < 0");
	yy_delete_buffer(YY_CURRENT_BUFFER);
	popincfile(&infile, &tagin, fd, &bs, &pid, &domacros, &lineno);
	/* linux possibly chokes on this
	FREE(cfgfileprev);
	 */
	cfgfileprev = cfgfilecurr;
	cfgfilecurr = infile;
	yy_switch_to_buffer(bs);
	if (incnum == 0)
		BEGIN(INITIAL);
	return;
}

/* push current input buffer info onto the file stack
 */
void
pushincfile (char *file, FILE *yyin, int *fd, YY_BUFFER_STATE bs, pid_t pid, int domcr, int lno)
{

	incfiles[incnum].file = strdup(file);
	incfiles[incnum].yyin = yyin;
	incfiles[incnum].fd[0] = *fd;
	incfiles[incnum].fd[1] = *(fd+1);
	incfiles[incnum].bs = bs;
	incfiles[incnum].pid = pid;
	incfiles[incnum].domcr = domcr;
	incfiles[incnum].lineno = lno;
	return;
}

/* pop input buffer info off the file stack
 */
void
popincfile (char **file, FILE **yyin, int *fdp, YY_BUFFER_STATE *bsp, pid_t *pidp, int *dmp, int *lnp)
{

	*file = incfiles[incnum].file;
	*yyin = incfiles[incnum].yyin;
	*fdp = incfiles[incnum].fd[0];
	*(fdp+1) = incfiles[incnum].fd[0];
	*bsp = incfiles[incnum].bs;
	*pidp = incfiles[incnum].pid;
	*dmp = incfiles[incnum].domcr;
	*lnp = incfiles[incnum].lineno;
	return;
}
