%{

/*
   alt.l -- PIKT alarm script lexer

   $Header: /home/cvs/ucgsb/pikt/src/piktd/alt.l,v 1.13 1999/10/04 05:22:50 harlan Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "../lib/common.h"
#include "piktd.h"
#include "alt.tab.h"

#define RETURN(token) altlval.sval = strdup(yytext); return token

#define ISVAR(c)	strchr("$#%@",c) && (isalnum((int)*(cp+1)) || (*(cp+1) == '\[')) && !escape

#define LP	40
#define RP	41
#define LB	91
#define RB	93

int ii;
int d;
char cc;
char *yyt;
int yyl;

int context;	/* STRING/NUMBER */

char outstr[2+2*(MAXLINE+1)];

int fn = 0;

int condparen(char *);
int strfn(char *);
int decompose(char *, char *);

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%x CS DS SS FS

%option noyywrap

%%

<*>^#.+\n			{ alterror("syntax error");
				}

<*>^{L}[-_\.0-9a-zA-Z]*		{ context = NIL;
				  BEGIN(INITIAL);
				  RETURN(_STANZA);
				}

<INITIAL,DS,SS>status		{ return _STATUS; }
<INITIAL,DS,SS>active		{ return _ACTIVE; }
<INITIAL,DS,SS>inactive		{ return _INACTIVE; }

<INITIAL,DS,SS>level		{ return _LEVEL; }
<INITIAL,DS,SS>emerg|emergency	{ return _EMERG; }
<INITIAL,DS,SS>alert|urgent	{ return _ALERT; }
<INITIAL,DS,SS>crit|critical	{ return _CRIT; }
<INITIAL,DS,SS>err|error	{ return _ERR; }
<INITIAL,DS,SS>warn|warning	{ return _WARNING; }
<INITIAL,DS,SS>notice		{ return _NOTICE; }
<INITIAL,DS,SS>info		{ return _INFO; }
<INITIAL,DS,SS>debug		{ return _DEBUG; }

<INITIAL,DS,SS>task		{ return _TASK; }

<INITIAL,DS,SS>input		{ return _INPUT; }
<INITIAL,DS,SS>proc|process	{ BEGIN(CS);
				  return _PROCESS;
				}
<INITIAL,DS,SS>file		{ return _REGFILE; }
<INITIAL,DS,SS>logfile		{ return _LOGFILE; }

<INITIAL,DS,SS>filter		{ BEGIN(CS);
				  return _FILTER;
				}

<INITIAL,DS,SS>seps|separators	{ return _SEPARATORS; }
<INITIAL,DS,SS>dat|data		{ BEGIN(DS);
				  return _DAT;
				}
<INITIAL,DS,SS>keys		{ BEGIN(DS);
				  return _KEYS;
				}

<INITIAL,DS,SS>init		{ return _INIT; }
<INITIAL,DS,SS,CS>begin		{ BEGIN(INITIAL);
				  return _BEGIN;
				}
<INITIAL,DS,SS,CS>end		{ BEGIN(INITIAL);
				  return _END;
				}
<INITIAL,DS,SS,CS>rule		{ BEGIN(INITIAL);
				  return _RULE;
				}

<INITIAL,DS,SS>do		{ context = NIL;
				  return _DO;
				}
<INITIAL,DS,SS>if		{ context = NUMBER;
				  return _IF;
				}
<INITIAL,DS,SS>if{W}*"("	{ context = NUMBER;
				  unput('[');
				  return _IF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif) {
				  context = NUMBER;
				  return _ELIF;
				}
<INITIAL,DS,SS>(elif|elsif|elseif){W}*"(" {
				  context = NUMBER;
				  unput('[');
				  return _ELIF;
				}
<INITIAL,DS,SS>else		{ context = NIL;
				  return _ELSE;
				}
<INITIAL,DS,SS>endif|fi		{ context = NIL;
				  return _ENDIF;
				}
<INITIAL,DS,SS>while		{ context = NUMBER;
				  return _WHILE;
				}
<INITIAL,DS,SS>while{W}*"("	{ context = NUMBER;
				  unput('[');
				  return _WHILE;
				}
<INITIAL,DS,SS>endwhile		{ context = NIL;
				  return _ENDWHILE;
				}
<INITIAL,DS,SS>repeat		{ context = NIL;
				  return _REPEAT;
				}
<INITIAL,DS,SS>until		{ context = NUMBER;
				  return _UNTIL;
				}
<INITIAL,DS,SS>until{W}*"("	{ context = NUMBER;
				  unput('[');
				  return _UNTIL;
				}
<INITIAL,DS,SS>for		{ context = NUMBER;
				  return _FOR;
				}
<INITIAL,DS,SS>for{W}*"("	{ context = NUMBER;
				  return _FORLP;
				}
<INITIAL,DS,SS>endfor		{ context = NIL;
				  return _ENDFOR;
				}
<INITIAL,DS,SS>break		{ context = NIL;
				  return _BREAK;
				}
<INITIAL,DS,SS>continue|cont	{ context = NIL;
				  return _CONTINUE;
				}
<INITIAL,DS,SS>switch		{ context = NIL;
				  return _SWITCH;
				}
<INITIAL,DS,SS>switch{W}*"("	{ context = NIL;
				  return _SWITCHLP;
				}
<INITIAL,DS,SS>case		{ context = NIL;
				  return _CASE;
				}
<INITIAL,DS,SS>default		{ context = NIL;
				  return _DEFAULT;
				}
<INITIAL,DS,SS>breakswitch|breaksw {
				  context = NIL;
				  return _BREAKSWITCH;
				}
<INITIAL,DS,SS>endswitch|endsw	{ context = NIL;
				  return _ENDSWITCH;
				}
<INITIAL,DS,SS>again		{ context = NIL;
				  return _AGAIN;
				}
<INITIAL,DS,SS>leave		{ context = NIL;
				  return _LEAVE;
				}
<INITIAL,DS,SS>redo		{ context = NIL;
				  return _REDO;
				}
<INITIAL,DS,SS>next		{ context = NIL;
				  return _NEXT;
				}
<INITIAL,DS,SS>skip		{ context = NUMBER;
				  return _SKIP;
				}
<INITIAL,DS,SS>last		{ context = NIL;
				  return _LAST;
				}
<INITIAL,DS,SS>pause		{ context = NUMBER;
				  return _PAUSE;
				}
<INITIAL,DS,SS>quit		{ context = NIL;
				  return _QUIT;
				}
<INITIAL,DS,SS>die		{ context = STRING;
				  return _DIE;
				}

<INITIAL,DS,SS>output		{ context = STRING;
				  return _OUTPUT;
				}
<INITIAL,DS,SS>mail		{ return _MAIL; }
<INITIAL,DS,SS>print		{ return _PRINT; }
<INITIAL,DS,SS>log		{ return _LOG; }
<INITIAL,DS,SS>piktlog		{ return _PIKTLOG; }
<INITIAL,DS,SS>syslog		{ return _SYSLOG; }

<INITIAL,DS,SS>set		{ context = NIL;
				  BEGIN(SS);
				  return _SET;
				}

<INITIAL,DS,SS>exec		{ context = STRING;
				  BEGIN(INITIAL);
				  return _EXEC;
				}
<INITIAL>wait			{ return _WAIT; }

<CS,DS>\"[^\n]+\"/[\n\t ]*	{ yytext[yyleng-1] = '\0';
				  BEGIN(INITIAL);
				  altlval.sval = strdup(yytext+1);
				  return _STR;
				}

<DS>({D}{1,5}|$|$-{D}{1,5})/[^\]a-zA-Z0-9] {
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>"["({D}{1,5}|$|$-{D}{1,5})"]" {
				  BEGIN(INITIAL);
				  RETURN(_POSITION);
				}
<DS>{D}+","({D}+|$)		{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}
<DS>"["{D}+","({D}+|$)"]"	{ BEGIN(INITIAL);
				  RETURN(_RANGE);
				}

<INITIAL,SS>"="			{ context = NIL;
				  BEGIN(INITIAL);
				  return yytext[0];
				}

<INITIAL,SS>(#|$){L}{DL}*({W}*|"["[^\n]+"]"{W}*)("+="|"-="|"*="|"/="|"%="|"\*\*="|"<<="|">>="|"&="|"|="|"^="|".=") {
				  if (!strcmp(".=",yytext))
					context = STRING;
				  else
					context = NUMBER;
				  yyt = strdup(yytext);
				  yyl = yyleng;
				  ii = yyl - 2;
				  while (ii >= 0)
					unput(*(yyt+ii--));
				  unput(' ');
				  unput('=');
				  unput(' ');
				  cc = *(yyt+yyl-3);
				  ii = yyl - ((isspace((int)cc) || isalnum((int)cc)) ? 3 : 4);
				  while (ii >= 0)
					unput(*(yyt+ii--));
				  free(yyt);
				}

<INITIAL>\"[^\"]*\"		{ if ((d = decompose(yytext, outstr)) > 0) {
					for (ii=d; ii>=0; ii--)
						unput(*(outstr+ii));
				  } else if (d < 0) {
					*(outstr-d) = '\0';
					altlval.sval = strdup(outstr+1);
					return _STR;
				  } else {
					*(outstr+strlen(outstr)-1) = '\0';
					altlval.sval = strdup(outstr+1);
					return _STRVAR;
				  }
				}

<SS,FS,INITIAL>\'[^\']*\'	{ yytext[yyleng-1] = '\0';
				  altlval.sval = strdup(yytext+1);
				  return _STR;
				}

<FS>[A-Z][A-Z0-9]*		{ sprintf(outstr, "?%s", yytext);
				  altlval.sval = strdup(outstr);
				  return _FLHNDL;
				}

<DS,SS,FS,INITIAL>(#|$){L}{DL}*	{ context = (yytext[0] == '#' ? NUMBER : STRING);
				  RETURN(yytext[0] == '#' ? _VARNUM : _VARSTR);
				}

<FS,INITIAL>(%|@){L}{DL}*/("["[^\[\]]+"]")*{W}*("+"|"-"|"*"|"/"|"%"|"**"|"& "|"| "|"^"|"<<"|">>"|"=="|">"|"<"|">="|"<="|"!>"|"!<"|"!="|"<>"|"]"){W}* {
				  context = NUMBER;
				  RETURN(yytext[0] == '%' ? _VARHSTNUM : _VARPRCNUM);
				}

<FS,INITIAL>(%|@){L}{DL}*/("["[^\[\]]+"]")*{W}*("."|"=~"|"=~~"|"!~"|"!~~"){W}* {
				  context = STRING;
				  RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<FS,INITIAL>(%|@){L}{DL}*/("["[^\[\]]+"]")*{W}+(eq|ne|lt|gt|le|ge){W}* {
				  context = STRING;
				  RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
				}

<FS,INITIAL>(%|@){L}{DL}*	{ if (!prescan) {
					switch (vardattype(yytext)) {
						case STRING:
							RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
							break;
						case NUMBER:
							RETURN(yytext[0] == '%' ? _VARHSTNUM : _VARPRCNUM);
							break;
						case ERROR:
						default:
							altlval.sval = strdup(err);
							return _STR;
							break;
					}
				  } else {
					switch (context) {
						case NUMBER:
							RETURN(yytext[0] == '%' ? _VARHSTNUM : _VARPRCNUM);
							break;
						case STRING:
						case NIL:
						default:
							RETURN(yytext[0] == '%' ? _VARHSTSTR : _VARPRCSTR);
							break;
					}
				  }
				}

<SS,FS,INITIAL>${D}{1,5}	{ context = STRING;
				  RETURN(_VALSTR);
				}

<SS,FS,INITIAL>$"["		{ yytext[yyleng-1] = '\0';
				  altlval.sval = strdup(yytext);
				  unput('[');
				  unput('[');
				  unput('[');
				  context = STRING;
				  return _VALSTR;
				}

<DS>(#|$){L}{DL}*"["		{ yytext[yyleng-1] = '\0';
				  altlval.sval = strdup(yytext);
				  unput('[');
				  unput('[');
				  unput('[');
				  context = (altlval.sval[0] == '#' ? NUMBER : STRING);
				  return (altlval.sval[0] == '#' ? _VARNUM : _VARSTR);
				}

<DS,SS,FS,INITIAL>"["		{ context = NUMBER;
				  return yytext[0];
				}
<DS,SS,FS,INITIAL>"[[["		{ return '['; }
<DS,SS,FS,INITIAL>{D}{1,5}	{ context = NUMBER;
				  RETURN(_INT);
				}
<DS,SS,FS,INITIAL>"]"		{ return yytext[0]; }

<SS,FS,INITIAL>("+"|"-"|"~")	{ context = NUMBER;
				  return yytext[0];
				}

<SS,FS,INITIAL>("*"|"/"|"%"|"**"|"&"|"|"|"^"|"<<"|">>") {
				  context = NUMBER;
				  RETURN(_NOP);
				}

<SS,FS,INITIAL>"."		{ context = STRING;
				  RETURN(_SOP);
				}

<SS,FS,INITIAL>(-e|-z|-f|-d|-c|-b|-p|-l|-S) {
				  context = STRING;
				  RETURN(_FOP);
				}

<SS,FS,INITIAL>("=="|">"|"<"|">="|"<="|"!>"|"!<"|"!="|"<>") {
				  context = NUMBER;
				  RETURN(_NREL);
				}
<SS,FS,INITIAL>("=~"|"=~~"|"!~"|"!~~"|eq|ne|lt|gt|le|ge) {
				  context = STRING;
				  RETURN(_SREL);
				}

<SS,FS,INITIAL>{D}{1,2}		{ context = NUMBER;
				  RETURN(_NUM);
				}
<SS,FS,INITIAL>(({D}+)|({D}*\.{D}+)|((({D}+)|({D}*\.{D}+))\/(({D}+)|({D}*\.{D}+))))%* {
				  context = NUMBER;
				  RETURN(_NUM);
				}

<SS,FS,INITIAL>(#|$){L}{DL}*/{W}*"(" {
				  context = (yytext[0] == '#' ? NUMBER : STRING);
				  BEGIN(FS);
				  RETURN(yytext[0] == '#' ? _FNNUM : _FNSTR);
				}
<SS,FS>"("			{ context = NIL;
				  fn++;
				  return '(';
				}
<SS,FS>")"			{ context = NIL;
				  fn--;
				  if (!fn)
					BEGIN(INITIAL);
				  return ')';
				}

<SS,FS>\"[^\n]*\"/[\n\t ]*	{ context = STRING;
				  ii = strfn(yytext);
				  yyless(ii+1);
				  yytext[ii+1] = '\0';
				  if ((d = decompose(yytext, outstr)) > 0) {
					for (ii=d; ii>=0; ii--)
						unput(*(outstr+ii));
				  } else if (d < 0) {
					*(outstr-d) = '\0';
					altlval.sval = strdup(outstr+1);
					return _STR;
				  } else {
					*(outstr+strlen(outstr)-1) = '\0';
					altlval.sval = strdup(outstr+1);
					return _STRVAR;
				  }
				}

"("[\n\t ]*[^\n\t ]+[\n\t ]*(=~|=~~|!~|!~~|eq|ne|lt|gt|le|ge)[\n\t ]*[^\n\t ]+[\n\t ]*")"[\n\t ]* { 
				  context = STRING;
				  yyless(1);
				  return '[';
				}

"("([-+*/%$#@~^><)(.0-9a-zA-Z\n\t\[\] ]|"&"[^&]|"|"[^|])+ {
				  if (condparen(yytext)) {
					context = NUMBER;
					yyless(1);
					return '[';
				  } else {
					context = NUMBER;
					yyless(1);
					return yytext[0];
				  }
				}

")"				{ return yytext[0]; }

<SS,FS,INITIAL>"!"		{ context = NUMBER;
				  return _NOT;
				}
<SS,FS,INITIAL>"&&"		{ context = NUMBER;
				  return _AND;
				}
<SS,FS,INITIAL>"||"		{ context = NUMBER;
				  return _OR;
				}

<*>","				{ return yytext[0]; }

<*>{W}				{ ; }
<*>\n				{ lineno++; }

<*>.				{ return yytext[0]; }

%%

/* alt error handler
 */
void
alterror (const char *message)
{
	static int l = 0;

	if (l != lineno)
		l = lineno;
	logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in .alt file, at or near line %d, %s (in altparse()): @%s@", lineno, message, yytext);
	BEGIN(INITIAL);
	return;
}

/* return whether the leading parenthesis in the supplied string is a
   logic conditional parenthesis, as opposed to an arithmetic parenthesis;
   basically, we count the number of left versus right parentheses, until
   we reach certain turning points, where if the number of counted left
   parentheses exceeds the right count, we know that we are entering a
   logical condition
 */
int
condparen (char *s)
{
	int i;
	int lp, rp;
	unsigned char c, nc;
	int done;

	done = FALSE;
	lp = rp = 0;
	i = 0;
	while ((c = *(s+i))) {
		switch(c) {
			case '(':
				lp++;
				break;
			case ')':
				rp++;
				if (lp == rp) {
					if (context == STRING)
						rp--;	/* signalling cond */
					done = TRUE;
				}
				break;
			case '<':
				nc = *(s+i+1);
				switch(nc) {
					case ' ':
					case '>':
					case '\n':
					case '\t':
						done = TRUE;
						break;
					default:
						break;
				}
				break;
			case '>':
				nc = *(s+i+1);
				switch(nc) {
					case ' ':
					case '\n':
					case '\t':
						done = TRUE;
						break;
					default:
						break;
				}
				break;
			case '-':	/* e.g., (-e $name) */
				nc = *(s+i+1);
				if (strchr("ezfdcbplS",nc) && (lp == 1))
					done = TRUE;
				break;
			default:
				break;
		}
		if (done) break;
		i++;
	}
	return lp > rp ? TRUE : FALSE;
}

/* when processing a statement like

      $substitute($line, "t", "X", 2) eq "this is a Xest"

   the lexer will eventually pattern match

      "t", "X", 2) eq "this is a Xest"

   this function isolates the next function string, i.e.,

      "t"

   returning its length, in this example, 3
 */
int
strfn (char *s)
{
	int i, j;
	int q;
	unsigned char c;
	int done;

	done = FALSE;
	q = 0;
	i = 0;
	while ((c = *(s+i))) {
		switch(c) {
			case '"':
				if (*(s+i+1) == '\0')
					done = TRUE;
				else
					q++;
				break;
			case '.':
			case ',':
			case ')':
				if (even(q)) {
					done = TRUE;
					j = 1;
					while ((c = *(s+i-j)) != '\"') {
						if (isspace(c))
							j++;
						else {
							done = FALSE;
							break;
						}
					}
					if (done)
						i -= j;
				}
				break;
			default:
				break;
		}
		if (done) break;
		i++;
	}
	return i;
}

/* given a string, break it down into its component parts--pure text strings,
   surrounded by single quotes, and variables or functions, joined together
   by the concatenation operator . (dot);

   for example, the input string

      "User $uname has uid $uid > the 60000 upper limit"

   when decomposed becomes

      'User ' . $uname . ' has uid ' . $uid . ' > the 60000 upper limit'

   the function returns the length of the output string

   the decomposition output string is pushed back into the lexer input
   stream for subsequent reprocessing
 */
int
decompose (char *istr, char *ostr)
{
	char *cp;
	int i;
	char c;
	int isvar;
	int escape;
	int brkt, prn;
	int decomposed;

	decomposed = FALSE;
	cp = istr;
	isvar = FALSE;
	escape = FALSE;
	brkt = prn = 0;
	i = 0;
	while ((c = *cp)) {
		if ((c == '\\') && !escape)  {
			escape = TRUE;
			cp++;
			continue;
		}
		if (isvar) {
			if ((c == LB) && !prn && !escape) {
				brkt++;
				*(ostr+i++) = c;
			} else if ((c == RB) && !prn && brkt && !escape) {
				brkt--;
				*(ostr+i++) = c;
			} else if ((c == LP) && !brkt && !escape) {
				prn++;
				*(ostr+i++) = c;
			} else if ((c == RP) && !brkt && prn && !escape) {
				prn--;
				*(ostr+i++) = c;
			} else if (isalnum((int)c) || ((brkt > 0) && !prn) || ((prn >0) && !brkt)) {
				*(ostr+i++) = c;
			} else {
				if (ISVAR(c)) {
					*(ostr+i) = '\0';
					strcat(ostr, " . ");
					i += 3;
					*(ostr+i++) = c;
				} else if ((c != '\"') || *(cp+1)) {
					*(ostr+i) = '\0';
					strcat(ostr, " . \'");
					i += 4;
					if (c == '\"') 
						*(ostr+i++) = DQTPROXY;
					else if (c == '\'') 
						*(ostr+i++) = SQTPROXY;
					else
						*(ostr+i++) = c;
					isvar = FALSE;
				}
			}
		} else { 
			if (ISVAR(c)) {
				if ((cp != istr+1) || ((cp == istr+1) && (c != '$'))) {
					*(ostr+i) = '\0';
					strcat(ostr, "\' . \'");
					i += 5;
				}
				i--;
				*(ostr+i++) = c;
				isvar = TRUE;
				decomposed = TRUE;
			} else if (c == '\"') {
				if ((cp == istr) || !*(cp+1))	/* beginning or end of istr */
					*(ostr+i++) = '\'';
				else
					*(ostr+i++) = DQTPROXY;
			} else if (c == '\'') {
				*(ostr+i++) = SQTPROXY;
			} else {
				*(ostr+i++) = c;
			}
		}
		escape = FALSE;
		cp++;
	}
	*(ostr+i) = '\0';
	/* the following if is for the case where the istr is, e.g.,
	   "\$trim", and where regtok would (mistakenly) detect an
	   attempt to use a reserved function name as a var name;
	 */
	if (*(istr+1) == '\\') {
		if (strchr("$#%@",*(ostr+1)) && !strncmp(istr+2, ostr+1, (size_t)(i-2))) {
			return 0;
		}
	}
	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "istr is %s, ostr is %s, decomposed is %d, i is %d", istr, ostr, decomposed, i);
	if (decomposed)
		return i-1;
	else
		return 1-i;
}
