/*
   debug.c -- PIKT pikt & piktd debug routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/debug.c,v 1.9 1999/10/04 05:22:50 harlan Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "../lib/common.h"
#include "piktd.h"

static const char *statements[] = {
				"",
				"DO",
				"IF",
				"ELIF",
				"ELSE",
				"ENDIF",
				"WHILE",
				"ENDWHILE",
				"REPEAT",
				"UNTIL",
				"FOR",
				"ENDFOR",
				"BREAK",
				"CONTINUE",
				"SWITCH",
				"CASE",
				"DEFAULT",
				"BREAKSWITCH",
				"ENDSWITCH",
				"AGAIN",
				"LEAVE",
				"REDO",
				"NEXT",
				"SKIP",
				"LAST",
				"PAUSE",
				"QUIT",
				"DIE",
				"OUTPUT",
				"MAIL",
				"PRINT",
				"LOCLOG",
				"PIKTLOG",
				"SYSLOG",
				"SET",
				"EXEC",
				"EXECWAIT",	/* 36 */
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"",
				"WHILEAUX",	/* 107 */
				"",
				"REPEATAUX",	/* 109 */
				"",
				"FORAUX"	/* 111 */
			      };

/* dump to stdout all vals for the current input line
 */
void
dumpvalnows (val *values)
{
	register int i, j, k;
	int maxsubary;
	int maxsubndx;

	printf("DEBUG: ");
	for (i=1; i<=MAXVARS; i++) {
		if (VARNAM(i) && (VALTYP(i) & FIXEDVAL)) {
			switch (VARTYP(i)) {
				case SCALAR:
					if (valnow->vals[i].s) {
						if (valnow->VALSCL(i,0))
							printf("%s = @%s@, ", VARNAM(i), valnow->VALSCL(i,0));
					}
					break;
				case ARRAY:
					if (valnow->vals[i].a) {
						maxsubary = SA(MAXARNDX(i));
						for (j=0; j<=maxsubary; j++) {
							if (valnow->vals[i].a->subary[j]) {
								if (j == maxsubary)
									maxsubndx = SN(MAXARNDX(i));
								else
									maxsubndx = 255;
								for (k=0; k<=maxsubndx; k++) {
									if (valnow->vals[i].a->subary[j]->val[k])
										printf("%s[%d] = @%s@, ", VARNAM(i), j, valnow->vals[i].a->subary[j]->val[k]);
								}
							}
						}
					}
					break;
				default:
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in dumpvalnow(), %s, unknown vartype for var %d", ALMNAM, i);
					STATUS = INACTIVE;
					break;
			}
		}
	}
	printf("\n");
	return;
}

/* dump to logreturn() all vals
 */
void
dumpvals(val *values)
{
	val *vlp;
	register int i, j, k;
	int maxsubary;
	int maxsubndx;

return;
	foreach (vlp, values) {
		for (i=1; i<=MAXVARS; i++) {
			switch (VARTYP(i)) {
				case SCALAR:
					if (vlp->vals[i].s) {
						if (vlp->VALSCL(i,0))
							logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "%s", vlp->VALSCL(i,0));
					}
					break;
				case ARRAY:
					if (vlp->vals[i].a) {
						maxsubary = SA(MAXARNDX(i));
						for (j=0; j<=maxsubary; j++) {
							if (vlp->vals[i].a->subary[j]) {
								if (j == maxsubary)
									maxsubndx = SN(MAXARNDX(i));
								else
									maxsubndx = 255;
								for (k=0; k<=maxsubndx; k++) {
									if (vlp->vals[i].a->subary[j]->val[k])
										logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "%s", vlp->vals[i].a->subary[j]->val[k]);
								}
							}
						}
					}
					break;
				default:
					break;
			}
		}
	}
	return;
}

/* dump linking info for all program statements
 */
void
dumpsps (stm *sps, char *message)
{
	stm *sp;

	logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "\n\n%-30s statement   type    sp tnext fnext   aux  next", message);
	foreach(sp, sps) {
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "               %-12s %3d %5x %5x %5x %5x %5x", statements[sp->type], sp->type, sp, sp->tnext, sp->fnext, sp->aux, sp->next);
	}
	fprintf(stderr, "\n");
	fflush(stderr);
}

/*
void
dumpalarms ()
{
	return;
}
 */

/*
void
dumptime (char *timestr, unsigned long time)
{
	register int i;
	unsigned long mask;

	printf("%s\t", timestr);
	for (i=31; i>=0; i--) {
		mask = (1 << i);
		printf("%c", time & mask ? '1' : '0');
		if (!(i % 8))
			printf(" ");
	}
	printf("\n");
}
 */
