%{

/*
   expr.l -- PIKT script expression lexer

   $Header: /home/cvs/ucgsb/pikt/src/piktd/expr.l,v 1.17 1999/10/17 06:02:19 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#include "../lib/common.h"
#include "piktd.h"
#include "expr.tab.h"

#define EXPERR		if (doeval && !iserrdef) exprres = ERROR;\
			cp = err

#define EXPNIL		if (doeval && !iserrdef) exprres = NIL;\
			cp = nil

#define NULSTR		tokcurr = tokcurr->next;\
			cp = (char *) NULL;\
			return '\0'

char *cp;

char arytypvrndx[16];
char vrndx[3];

int exprinput(void);

/* note: inlin, inline & inputline should all be absent from these lists! */

char fnnumnames[] = " abs absval absolutevalue age and ascii average avg ceiling datevalue day daynumber daysbetween defined endofinput eoi err error even false fclose fileage filemode fileino filelinks fileuid filegid filesize fileatime filemtime filectime find fopen grgid groupgid hour if index inputlength inlen inputnumber innum int integer isalnum isalpha isblank iscntrl isdigit isgraph islower isprint ispunct isspace isupper length max maximum mean median min minimum minute mode month monthnumber nil not now odd or ord parse pclose popen pid procid processid pprocid pprocessid ppid pwuid passwduid pwgid passwdgid read round rfind rindex rsearch search second system timevalue today true trunc truncate val value weekday write year yearday yearweek ";

char fnstrnames[] = " alarm ampm basename char chop command dayname dirname dquote err fixed grname groupname grpassword grouppassword grmem groupmem grmembers groupmembers hostname if left lower ltrim mid middle monthname newline nil proper pwname passwdname pwpassword passwdpassword pwcomment passwdcomment pwgecos passwdgecos pwdir passwddir pwshell passwdshell repeat replace reverse right rtrim script space squote substitute substring substr str string tab text trim upper ";

#define RETURN(token) exprlval.sval = strdup(yytext); return token

#undef YY_INPUT
#define YY_INPUT(buf, result, max_size) { \
		char c = exprinput(); \
		result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
	}

%}

W		[ \t]
D		[0-9]
L		[A-Za-z]
DL		[0-9A-Za-z]

%option noyywrap

%%

"("				{ RETURN(_LP); }
")"				{ RETURN(_RP); }
"["				{ RETURN(_LB); }
"]"				{ RETURN(_RB); }
\"\[\"				{ RETURN(_QLB); }
","				{ RETURN(_CMA); }
"!"				{ RETURN(_NOT); }
"&&"				{ RETURN(_AND); }
"||"				{ RETURN(_OR); }

"+"				{ RETURN(_ADD); }
"-"				{ RETURN(_SUBTRACT); }
"*"				{ RETURN(_MULTIPLY); }
"/"				{ RETURN(_DIVIDE); }
"%"				{ RETURN(_MODULUS); }
"**"				{ RETURN(_EXPONENTIATE); }
"&"				{ RETURN(_BITAND); }
"|"				{ RETURN(_BITOR); }
"^"				{ RETURN(_BITEXOR); }
"~"				{ RETURN(_BITCOMP); }
"<<"				{ RETURN(_BITLSHFT); }
">>"				{ RETURN(_BITRSHFT); }

"."				{ RETURN(_CONCATENATE); }

"=="				{ RETURN(_EQ); }
">"				{ RETURN(_GT); }
"<"				{ RETURN(_LT); }
">="				{ RETURN(_GE); }
"<="				{ RETURN(_LE); }
"!>"				{ RETURN(_NGT); }
"!<"				{ RETURN(_NLT); }
"!="				{ RETURN(_NE); }
"<>"				{ RETURN(_NE2); }
"=~"				{ RETURN(_RGXP); }
"=~~"				{ RETURN(_RGXPI); }
"!~"				{ RETURN(_NRGXP); }
"!~~"				{ RETURN(_NRGXPI); }
eq				{ RETURN(_SEQ); }
ne				{ RETURN(_SNE); }
lt				{ RETURN(_SLT); }
gt				{ RETURN(_SGT); }
le				{ RETURN(_SLE); }
ge				{ RETURN(_SGE); }

-e				{ RETURN(_EXIST); }
-z				{ RETURN(_ZEROLEN); }
-f				{ RETURN(_REGFILE); }
-d				{ RETURN(_DIRECTORY); }
-c				{ RETURN(_CHARSPEC); }
-b				{ RETURN(_BLKSPEC); }
-p				{ RETURN(_PIPE); }
-l				{ RETURN(_SYMLINK); }
-S				{ RETURN(_SOCKET); }

#abs|#absval|#absolutevalue	{ RETURN(_ABSVAL); }
#age|#fileage			{ RETURN(_AGE); }
$alarm				{ RETURN(_ALARM); }
$ampm				{ RETURN(_AMPM); }
#and				{ RETURN(_ANDFN); }
#ascii				{ RETURN(_CODE); }
#avg|#average			{ RETURN(_AVERAGE); }
$basename			{ RETURN(_BASENAME); }
#ceiling			{ RETURN(_CEILING); }
$char				{ RETURN(_CHAR); }
$chop				{ RETURN(_CHOP); }
#code				{ RETURN(_CODE); }
$command			{ RETURN(_COMMAND); }
#datevalue			{ RETURN(_DATEVALUE); }
#day				{ RETURN(_DAY); }
$dayname			{ RETURN(_DAYNAME); }
#daynumber			{ RETURN(_DAYNUMBER); }
#daysbetween			{ RETURN(_DAYSBETWEEN); }
#defined			{ iserrdef = TRUE; RETURN(_DEFINED); }
$dirname			{ RETURN(_DIRNAME); }
$dquote				{ RETURN(_DQUOTE); }
#eoi|#endofinput		{ RETURN(_ENDOFINPUT); }
#err				{ RETURN(_ERRFNNUM); }
$err				{ RETURN(_ERRFNSTR); }
#error				{ iserrdef = TRUE; RETURN(_ERROR); }
#even				{ RETURN(_EVEN); }
#false				{ RETURN(_FALSEFN); }
#fclose				{ RETURN(_FCLOSE); }
#filemode			{ RETURN(_FILEMODE); }
#fileino			{ RETURN(_FILEINO); }
#filelinks			{ RETURN(_FILELINKS); }
#fileuid			{ RETURN(_FILEUID); }
#filegid			{ RETURN(_FILEGID); }
#filesize			{ RETURN(_FILESIZE); }
#fileatime			{ RETURN(_FILEATIME); }
#filemtime			{ RETURN(_FILEMTIME); }
#filectime			{ RETURN(_FILECTIME); }
#find				{ RETURN(_SEARCH); }
$fixed				{ RETURN(_FIXED); }
#fopen				{ RETURN(_FOPEN); }
$grname|$groupname		{ RETURN(_GRNAME); }
$grpassword|$grouppassword	{ RETURN(_GRPASSWORD); }
#grgid|#groupgid		{ RETURN(_GRGID); }
$grmem|$grmembers|$groupmem|$groupmembers { RETURN(_GRMEMBERS); }
$hostname			{ RETURN(_HOSTNAME); }
#hour				{ RETURN(_HOUR); }
#if				{ RETURN(_IFFNNUM); }
$if				{ RETURN(_IFFNSTR); }
#index				{ RETURN(_SEARCH); }
$inlin|$inline|$inputline	{ RETURN(_INPUTLINE); }
#inlen|#inputlength		{ RETURN(_INPUTLENGTH); }
#innum|#inputnumber		{ RETURN(_INPUTNUMBER); }
#int|#integer			{ RETURN(_INTEGER); }
#isalnum			{ RETURN(_ISITALNUM); }
#isalpha			{ RETURN(_ISITALPHA); }
#isblank			{ RETURN(_ISITBLANK); }
#iscntrl			{ RETURN(_ISITCNTRL); }
#isdigit			{ RETURN(_ISITDIGIT); }
#isgraph			{ RETURN(_ISITGRAPH); }
#islower			{ RETURN(_ISITLOWER); }
#isprint			{ RETURN(_ISITPRINT); }
#ispunct			{ RETURN(_ISITPUNCT); }
#isspace			{ RETURN(_ISITSPACE); }
#isupper			{ RETURN(_ISITUPPER); }
$left				{ RETURN(_LEFT); }
$ltrim				{ RETURN(_LTRIM); }
#length				{ RETURN(_LENGTH); }
$lower				{ RETURN(_LOWER); }
#max|#maximum			{ RETURN(_MAXIMUM); }
#mean				{ RETURN(_AVERAGE); }
#median				{ RETURN(_MEDIAN); }
$mid|$middle			{ RETURN(_SUBSTRING); }
#min|#minimum			{ RETURN(_MINIMUM); }
#minute				{ RETURN(_MINUTE); }
#mode				{ RETURN(_MODE); }
#month				{ RETURN(_MONTH); }
$monthname			{ RETURN(_MONTHNAME); }
#monthnumber			{ RETURN(_MONTHNUMBER); }
$newline			{ RETURN(_NEWLINE); }
#nil				{ RETURN(_NILFNNUM); }
$nil				{ RETURN(_NILFNSTR); }
#not				{ RETURN(_NOTFN); }
#now				{ RETURN(_NOW); }
#odd				{ RETURN(_ODD); }
#or				{ RETURN(_ORFN); }
#ord				{ RETURN(_CODE); }
#parse				{ RETURN(_PARSE); }
#pclose				{ RETURN(_PCLOSE); }
#popen				{ RETURN(_POPEN); }
#procid|#processid|#pid		{ RETURN(_PROCID); }
#pprocid|#pprocessid|#ppid	{ RETURN(_PPROCID); }
$proper				{ RETURN(_PROPER); }
$pwname|$passwdname		{ RETURN(_PWNAME); }
$pwpassword|$passwdpassword	{ RETURN(_PWPASSWORD); }
#pwuid|#passwduid		{ RETURN(_PWUID); }
#pwgid|#passwdgid		{ RETURN(_PWGID); }
$pwgecos|$pwcomment|$passwdgecos|$passwdcomment { RETURN(_PWGECOS); }
$pwdir|$passwddir		{ RETURN(_PWDIR); }
$pwshell|$passwdshell		{ RETURN(_PWSHELL); }
#read				{ RETURN(_READ); }
$repeat				{ RETURN(_REPEAT); }
$replace			{ RETURN(_REPLACE); }
$reverse			{ RETURN(_REVERSE); }
$right				{ RETURN(_RIGHT); }
#round				{ RETURN(_ROUND); }
#rfind|#rindex|#rsearch		{ RETURN(_RSEARCH); }
$rtrim				{ RETURN(_RTRIM); }
$script				{ RETURN(_ALARM); }
#search				{ RETURN(_SEARCH); }
#second				{ RETURN(_SECOND); }
$space				{ RETURN(_SPACE); }
$squote				{ RETURN(_SQUOTE); }
$substitute			{ RETURN(_SUBSTITUTE); }
$substr|$substring		{ RETURN(_SUBSTRING); }
$str|$string			{ RETURN(_FIXED); }
#system				{ RETURN(_SYSTEM); }
$tab				{ RETURN(_TAB); }
$text				{ RETURN(_FIXED); }
#timevalue			{ RETURN(_TIMEVALUE); }
#today				{ RETURN(_TODAY); }
$trim				{ RETURN(_TRIM); }
#true				{ RETURN(_TRUEFN); }
#trunc|#truncate		{ RETURN(_TRUNC); }
$upper				{ RETURN(_UPPER); }
#val|#value			{ RETURN(_VALUE); }
#weekday			{ RETURN(_WEEKDAY); }
#write				{ RETURN(_WRITE); }
#year				{ RETURN(_YEAR); }
#yearday			{ RETURN(_YEARDAY); }
#yearweek			{ RETURN(_YEARWEEK); }

"[[ARYSTR "{D}{1,5}"]]"		{ /* {D}{1,5} for 1 .. 65535 */
				  yytext[yyleng-2] = '\0';
				  exprlval.ival = atoi(yytext+9);
				  return _ARYSTR;
				}
"[[ARYNUM "{D}{1,5}"]]"		{ /* {D}{1,5} for 1 .. 65535 */
				  yytext[yyleng-2] = '\0';
				  exprlval.ival = atoi(yytext+9);
				  return _ARYNUM;
				}

0+{D}+/[\n\177]			{ exprlval.sval = strdup(yytext);
				  return _STR;
				}

-?{D}+/[\n\177]			{ dpls = 0;
				  exprlval.lval = atol(yytext);
				  return _INT;
				}
-?({D}*\.{D}+)/[\n\177]		{ cp = strchr(yytext,'.')+1;
				  dpls = 0;
				  while (*cp) {
					dpls++;
					cp++;
				  }
				  exprlval.fval = atof(yytext);
				  return _NUM;
				}
-?(({D}+)|({D}*\.{D}+))%	{ dpls = 2;
				  if ((cp = strchr(yytext,'.'))) {
					cp++;
				  	while (*cp) {
						dpls++;
						cp++;
					}
				  }
				  yytext[yyleng-1] = '\0';
				  exprlval.fval = atof(yytext)/100;
				  return _NUM;
				}
-?(({D}+)|({D}*\.{D}+))\/(({D}+)|({D}*\.{D}+)) {
				  cp = strchr(yytext,'/');
				  if (atof(cp+1)) {
					*cp = '\0';
					dpls = 6;	/* arbitrary, i know */
					exprlval.fval = atof(yytext)/atof(cp+1);
					return _NUM;
				  } else {
					logreturn(PIKTLOG, iserrdef ? LOG_WARNING : LOG_ERR, __FILE__, __LINE__, FALSE, "in %s, attempted divide by zero: %s", ALMNAM, yytext);
					return _ERR;
				  }
				}

\000				{ return _NULSTR; }

[^\177\"]+			{ exprlval.sval = strdup(yytext);
				  switch (yytext[0]) {
					case ERROR:
						return _ERR;
						break;
					case NIL:
						return _NIL;
						break;
					default:
						return _STR;
						break;
				  }
				}

\"				{ return _QT; }

(\177)+				{ ; }

%%

/* expr error handler
 */
void
exprerror (const char *message)
{

	/* the following kludge gets around the problem of blank lines
	   being passwd to the expr parser in a rule like the following:

	   $line regexp "?*" "$line"

	   when we want to see all (new) lines in a (log) file, for
	   example; need a better solution here
	 */
	if (strcmp(yytext, "?*")) {
		if (isdebug)
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in %s, in exprparse(), expr %s: @%s@%d@", ALMNAM, message, yytext, yytext[0]);
		else
			logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in %s, in exprparse(), expr %s: @%s@%c@", ALMNAM, message, yytext, yytext[0]);
	}
	if (doeval) exprres = ERROR;
	return;
}

/* for the current statement, feed its token list to the input stream;
   if we encounter a special var token, feed the appropriate val
 */
int
exprinput (void)
{
	static signed char c  = '\0';
	static signed char c1;
	static unsigned char c2;
	static char *cp = (char *) NULL;
	static unsigned offset = 0 ;
	static int ary = 0;

	if (!tokcurr) {
		return EOF;
	}

	/* set next cp */
	if (!cp) {
		offset = 0;
		if ((c = c1 = *tokcurr->token)) {
			if (c1 > 0) {
				cp = tokcurr->token;
			} else if (c1 < 0) {
				c2 = (unsigned char) *((tokcurr->token)+1);
				switch (c1) {
					case VLNOW:
						if ((DATTYP(c2) == FILEHANDLE) ||
						    (DATTYP(c2) == PROCHANDLE)   ) {
							sprintf(vrndx, "%d", c2);
							cp = vrndx;
						} else {
							switch (VARTYP(c2)) {
								case SCALAR:
									if ((cp = valnow->VALSCL(c2,0))) {
										if (!(c = *cp)) {
											NULSTR;
										}
									} else {
										logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, attempt to reference nonexistent value for var %s (%d)", ALMNAM, VARNAM(c2), c2);
										EXPNIL;
									}
									break;
								case ARRAY:
									if (vapndx > MAXVAP-1) {
										logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, max array index nesting (%d) exceeded", ALMNAM, MAXVAP);
										EXPERR;
									} else {
										vap[vapndx++] = valnow->vals[c2].a;
										sprintf(arytypvrndx, "[[%s %d]]", arytyp[DATTYP(c2)], c2);
										cp = arytypvrndx;
									}
									break;
							}
						}
						break;
					case VLHST:
						if (HSTERR || !valhst) {
							if (VARTYP(c2) == ARRAY) {
								tokcurr = tokcurr->next;
								do {
									if (tokcurr->token[0] == '\[')
										ary++;
									else if (tokcurr->token[0] == ']')
										ary--;
									if (ary)
										tokcurr = tokcurr->next;
								} while (ary);
							}
							if (isverboselog)
								logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, attempt to reference nonexistent history value for var %s%s (%d)", ALMNAM, VARNAM(c2), VARTYP(c2) == SCALAR ? "" : "[]", c2);
							EXPNIL;
						} else {
							switch (VARTYP(c2)) {
								case SCALAR:
									if (valhst->VALSCL(c2,0)) {
										cp = valhst->VALSCL(c2,0);
										if (!(c = *cp)) {
											NULSTR;
										}
									} else if (valglb->VALSCL(c2,0)) {
										cp = valglb->VALSCL(c2,0);
										if (!(c = *cp)) {
											NULSTR;
										}
									} else {
										if (isverboselog)
											logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, attempt to reference nonexistent history value for var %s (%d)", ALMNAM, VARNAM(c2), c2);
										EXPNIL;
									}
									break;
								case ARRAY:
									if (vapndx > MAXVAP-1) {
										logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, max array index nesting (%d) exceeded", ALMNAM, MAXVAP);
										EXPERR;
									} else {
										vap[vapndx++] = valhst->vals[c2].a;
										sprintf(arytypvrndx, "[[%s %d]]", arytyp[DATTYP(c2)], c2);
										cp = arytypvrndx;
									}
									break;
							}
						}
						break;
					case VLPRC:
						if (innum == 1) {
							if (VARTYP(c2) == ARRAY) {
								tokcurr = tokcurr->next;
								do {
									if (tokcurr->token[0] == '\[')
										ary++;
									else if (tokcurr->token[0] == ']')
										ary--;
									if (ary)
										tokcurr = tokcurr->next;
								} while (ary);
							}
							if (isverboselog)
								logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, attempt to reference nonexistent preceding value for var %s%s (%d)", ALMNAM, VARNAM(c2), VARTYP(c2) == SCALAR ? "" : "[]", c2);
							EXPNIL;
						} else {
							switch (VARTYP(c2)) {
								case SCALAR:
									if (valprc->VALSCL(c2,0)) {
										cp = valprc->VALSCL(c2,0);
										if (!(c = *cp))	{
											NULSTR;
										}
									} else {
										if (isverboselog)
											logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprintput(), %s, attempt to reference nonexistent preceding value for var %s (%d)", ALMNAM, VARNAM(c2), c2);
										EXPNIL;
									}
									break;
								case ARRAY:
									if (vapndx > MAXVAP-1) {
										logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, max array index nesting (%d) exceeded", ALMNAM, MAXVAP);
										EXPERR;
									} else {
										vap[vapndx++] = valprc->vals[c2].a;
										sprintf(arytypvrndx, "[[%s %d]]", arytyp[DATTYP(c2)], c2);
										cp = arytypvrndx;
									}
									break;
							}
						}
						break;
					case VLPRS:
						if (!c2 && (tokcurr->next && (*(tokcurr->next->token) == '\['))) {

							if (vapndx > MAXVAP-1) {
								logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, max array index nesting (%d) exceeded", ALMNAM, MAXVAP);
								EXPERR;
							} else {
								vap[vapndx++] = (chrary *) NULL;
								sprintf(arytypvrndx, "[[%s %d]]", arytyp[STRING], c2);
								cp = arytypvrndx;
							}
						} else if (c2 > matches) {
							logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, out-of-bounds val $%d", ALMNAM, c2);
							EXPERR;
						} else {
							cp = valscn[c2];
							if (!(c = *cp))	{
								NULSTR;
							}
						}
						break;
					case INLINNOW:
						if (!inlinnow) {
							EXPNIL;
						} else {
							cp = inlinnow;
							if (!(c = *cp))	{
								NULSTR;
							}
						}
						break;
					case INLINPRC:
						if (!inlinprc) {
							EXPNIL;
						} else {
							cp = inlinprc;
							if (!(c = *cp))	{
								NULSTR;
							}
						}
						break;
					case RDLINNOW:
						if (!rdlinnow) {
							EXPNIL;
						} else {
							cp = rdlinnow;
							if (!(c = *cp))	{
								NULSTR;
							}
						}
						break;
					case RDLINPRC:
						if (!rdlinprc) {
							EXPNIL;
						} else {
							cp = rdlinprc;
							if (!(c = *cp))	{
								NULSTR;
							}
						}
						break;
					case SQTPROXY:
						cp = tokcurr->token;
						break;
					default:
						logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, out-of-bounds var %d", ALMNAM, c1);
						EXPERR;
						break;
				}	/* end of switch (c1) */
			} else /* c1 == 0 */ {	/* should never be reached */
				logreturn(PIKTLOG, doeval ? LOG_ERR : LOG_WARNING, __FILE__, __LINE__, FALSE, "in exprinput(), %s, out-of-bounds var %d", ALMNAM, 0);
				EXPERR;
			}
		} else /* c1 == 0 */ {		/* null string */
			NULSTR;
		}
	}

	if (cp)
		c = *(cp+offset++);
	if (!c) {
		c = '\177';
		tokcurr = tokcurr->next;
		cp = (char *) NULL;
	}
	return c;

}
