/*
   lock.c -- PIKT pikt & piktd lock routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/lock.c,v 1.11 1999/10/04 05:22:51 harlan Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"
#include "piktd.h"

int islock(char *);
int mklock(char *, char *, int, int);
int rmlock(char *, int, int);
int rmlocks(int, int);

/* return TRUE if lockfile exists, FALSE if not
 */
int
islock (char *lockfile)
{
	struct stat fbuf;

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in islock(), looking for %s file", lockfile);
	if (lstat(lockfile, &fbuf))
		return FALSE;
	else
		return TRUE;
}

/* open a lock file, write contents if any, then close
 */
int
mklock (char *lockfile, char *contents, int interact, int log)
{
	FILE *lkfil;

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in mklock(), making %s file with contents: %s", lockfile, contents ? contents : "");
	if ((lkfil = fopen(lockfile, "w")) == NULL) {
		if (interact) {
			fprintf(stderr, "unable to create %s file!\n", lockfile);
			fflush(stderr);
		}
		if (log)
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in mklock(), fopen() error for %s", lockfile);
		return ERROR;
	}
	if (contents)
		fprintf(lkfil, "%s\n", contents);
	if (fclose(lkfil)) {
		if (interact) {
			fprintf(stderr, "unable to create %s file!\n", lockfile);
			fflush(stderr);
		}
		if (log)
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE,  "in mklock(), fclose() error for %s", lockfile);
		return ERROR;
	}
	chmod(lockfile, (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP));	/* 640 */
	return 0;
}

/* remove a lock file
 */
int
rmlock (char *lockfile, int interact, int log)
{

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in rmlock(), removing %s file", lockfile);
	if (remove(lockfile)) {		/* remove returns 0 on success */
		if (interact) {
			fprintf(stderr, "unable to remove %s file!\n", lockfile);
			fflush(stderr);
		}
		if (log)
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in rmlock(), remove() error for %s", lockfile);
		return ERROR;
	}
	return 0;
}

/* remove all pertinent locks
 */
int
rmlocks (int interact, int log)
{
	int res;

	res = 0;
	#ifdef PIKTD
	if (islock(piktdlock)) {
		if (rmlock(piktdlock, FALSE, log) == ERROR)
			res = ERROR;
	}
	#else
	if (islock(altlockfile)) {
		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in rmlocks(), removing %s file", altlockfile);
		if (rmlock(altlockfile, interact, log) == ERROR)
			res = ERROR;
	}
	/* remove any remaining exec lock */
	if (islock(execlockfile)) {
		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in rmlocks(), removing %s file", execlockfile);
		if (rmlock(execlockfile, interact, log) == ERROR)
			res = ERROR;
	}
	#endif
	return res;
}
