/*
   scan.c -- PIKT alarm input scanning routines

   $Header: /home/cvs/ucgsb/pikt/src/piktd/scan.c,v 1.19 1999/10/17 06:07:59 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-99 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
*/

#include "../lib/common.h"

#include "piktd.h"

#ifndef SEEK_SET
#define	SEEK_SET 0
#endif
#ifndef SEEK_CUR
#define	SEEK_CUR 1
#endif
#ifndef SEEK_END
#define	SEEK_END 2
#endif

char inlin[2][MAXLINE+1];
char *inlinnow, *inlinprc, *inlintmp;
int innum;
int inlen;
int skip;
int eoi;

char *valscn[MAXARYNDX+1];

int scanfields(char *, char *, int);
int scancolumns(char *);
int issep(int, char *);

/* execute the statements in the script begin section, if any;
   scan the input data from any rules: open, and start reading in, the
   input stream, if any; start writing to the history file as necessary;
   execute the statements in the script end section, if any;
   close the input stream, if any; finish writing to the history file
   as necessary
 */
int
scan (int inputtype, char *inputsource, int postype)
{
	FILE *infile;
	struct stat fbuf;
	ino_t inocurr;
	off_t sizecurr;
	unsigned long poscurr;
	unsigned long posend;
	char scancmd[MAXCMD+1];
	char *sc;
	int concat;
	char keystr[MAXLINE+1];
	ky *kp;
	val *vlp;
	register int i, j;
	char c;
	pid_t pid;

	inocurr = 0;
	sizecurr = 0;
	posend = 0;
	if (inputsource && (strlen(inputsource) > MAXCMD)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, input exceeds %d char limit", ALMNAM, MAXCMD);
		STATUS = INACTIVE;
		return ERROR;
	}
	if ((inputtype != NOINPUT) && !strlen(inputsource)) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, no input source", ALMNAM);
			STATUS = INACTIVE;
			return ERROR;
	}
	innum = 0;
	skip = 0;
	eoi = FALSE;
	infile = (FILE *) NULL;
	*inlin[0] = *inlin[1] = '\0';
	inlinnow = inlinprc = (char *) NULL;
	valhst = valglb;
	if (alarmcurr->begin) {
		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "BEGIN");
		stmcurr = alarmcurr->begin->statements;
		while (stmcurr) {
			switch (dostm(stmcurr->type)) {
				case TRUE:
					stmcurr = stmcurr->tnext;
					break;
				case FALSE:
					stmcurr = stmcurr->fnext;
					break;
				case AGAIN:
					stmcurr = alarmcurr->begin->statements;
					break;
				case QUIT:
					goto _QUIT;
					break;
				default:
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, invalid dostm() return at begin", ALMNAM);
					STATUS = INACTIVE;
					goto _QUIT;
					break;
			}
		}
	}
	switch (inputtype) {
		case NOINPUT:
			break;
		case PROCINPUT:
			sprintf(scancmd, "%s%s%s", inputsource, FILTER ? " | " : "", FILTER ? FILTER : "");
			sc = stripesc(scancmd);
			if ((infile = scanopen(sc, &pid)) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, scanopen() error for %s", ALMNAM, sc);
				STATUS = INACTIVE;
				return ERROR;
			}
			break;
		case LOGINPUT:
			if (stat(inputsource, &fbuf) != 0) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, stat() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			inocurr = fbuf.st_ino;
			sizecurr = fbuf.st_size;
			if ((infile = fopen(inputsource, "r")) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fopen() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			/* if this is the same log file as before,
			   and it has not shrunk (been zeroed out)
			 */
			if ((inocurr == (ino_t)INPUTLOGINO) && (sizecurr >= (off_t)INPUTLOGSIZE)) {
			/* but what about case where we have zeroed out
			   log file and it has since grown to be larger
			   than last alert run?  an unusual case, but it
			   could happen!
			 */
				/* if the previous file was not zero length,
				   attempt to position to previous eof pos
				 */
				if (INPUTLOGPOSEND) {
					if (fseek(infile, (long)INPUTLOGPOSEND, SEEK_SET)) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fseek() error for %s", ALMNAM, inputsource);
						rewind(infile);
						if (fclose(infile))
							logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for %s", ALMNAM, inputsource);
						STATUS = INACTIVE;
						return ERROR;
					}
				}
			}
			/* determine current position in file for passing
			   to tail command
			 */
			if ((poscurr = ftell(infile)) == -1L) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, ftell() error for %s", ALMNAM, inputsource);
				if (fclose(infile))
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			/* determine eof position for this file */
			if (fseek(infile,0L,SEEK_END)) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fseek() error for %s", ALMNAM, inputsource);
				if (fclose(infile))
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			} else {
				if ((posend = ftell(infile)) == -1L) {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, ftell() error for %s", ALMNAM, inputsource);
					if (fclose(infile))
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for %s", ALMNAM, inputsource);
					STATUS = INACTIVE;
					return ERROR;
				}
			}
			if (fclose(infile)) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			sprintf(scancmd, "%s +%ld%c %s%s%s", TAILCMD, poscurr+1, 'c', inputsource, FILTER ? " | " : "", FILTER ? FILTER : "");
			sc = stripesc(scancmd);
			if ((infile = scanopen(sc, &pid)) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, scanopen() error for %s", ALMNAM, sc);
				STATUS = INACTIVE;
				return ERROR;
			}
			break;
		case FILEINPUT:
			if (stat(inputsource, &fbuf) != 0) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, stat() error for %s", ALMNAM, inputsource);
				STATUS = INACTIVE;
				return ERROR;
			}
			sprintf(scancmd, "/bin/cat %s%s%s", inputsource, FILTER ? " | " : "", FILTER ? FILTER : "");
			sc = stripesc(scancmd);
			if ((infile = scanopen(sc, &pid)) == NULL) {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, scanopen() error for %s", ALMNAM, sc);
				STATUS = INACTIVE;
				return ERROR;
			}
			break;
	}
	/* write alarm name to history file */
	if (HSTTYP /* || KEYS || INPUTLOGNAME */)
		fprintf(hstout, "A %s\n", ALMNAM);
	if (HSTTYP & (FIXEDVAL|FREEVAL|BEGINVAL|ENDVAL) /* KEYS */) {
		for (i=1; i<=MAXVARS; i++) {
			if (NEWHST(i)) {
				if (VARNAM(i))
					fprintf(hstout, "R %d %d %d %d %s\n", i, VALTYP(i), VARTYP(i), DATTYP(i), VARNAM(i));
				else {
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, missing name for var %d", ALMNAM, i);
					if (fclose(infile))
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for %s", ALMNAM, inputsource);
					if (fclose(hstout))
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for %s", ALMNAM, hsttmpfile);
					STATUS = INACTIVE;
					return ERROR;
				}
			}
		}
		if (KEYS) {
			foreach (kp, KEYS)
				fprintf(hstout, "K %d %d\n", kp->vrndx, kp->arndx);
		}
	}
	inlinnow = inlin[0];
	if (inputtype != NOINPUT) {
		while (fgets(inlinnow, MAXLINE+1, infile) != NULL) {
			if (!alarmcurr->rules) {
				eoi = TRUE;
				break;
			}
			innum++;
			if (skip) {
				skip--;
				continue;
			}
			if ((c = fgetc(infile)) == EOF)
				eoi = TRUE;
			else
				ungetc(c, infile);
			inlen = dequote(inlinnow)-1;  /* subtract one for eol */
			if (!*inlinnow) continue;
			valtmp = valprc;
			valprc = valnow;
			valnow = valtmp;
			freevals(&valnow, FIXEDVAL|FREEVAL, FALSE);
			/* carry forward free vals to current line */
			for (i=1; i<=MAXVARS; i++) {
				if (VALTYP(i) & FREEVAL) {
					switch (VARTYP(i)) {
						case SCALAR:
							if (valprc->VALSCL(i,0))
								valnow->VALSCL(i,0) = strdup(valprc->VALSCL(i,0));
							break;
						case ARRAY:
							for (j=1; j<=MAXARNDX(i); j++) {
								if (valprc->SUBARY(i,j) && valprc->VALARY(i,j)) {
									if (! valnow->SUBARY(i,j))
										initchrsubary(&valnow->SUBARY(i,j));
									valnow->VALARY(i,j) = strdup(valprc->VALARY(i,j));
								}
							}
							break;
						case NIL:
							break;
						default:
							logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, unknown vartype for var %d", ALMNAM, i);
							STATUS = INACTIVE;
							goto _QUIT;
							break;
					}
				}
			}
			switch (postype) {
				case FIELDS:
					freevalscn();
					if ((matches = scanfields(inlinnow, SEPARATORS, FALSE)) != DATCNT) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, data mismatch in line: %s", ALMNAM, inlinnow);
						continue;
					}
					break;
				case COLUMNS:
					freevalscn();
					if ((matches = scancolumns(inlinnow)) != DATCNT) {
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, data mismatch in line: %s", ALMNAM, inlinnow);
						continue;
					}
					break;
				case REGEXPR:
					regexp(inlinnow, REGEXP, REG_EXTENDED, FALSE);
					break;
				default:	/* no dat statements,
						   postype defaults to 0
						 */
					freevalscn();
					matches = scanfields(inlinnow, SEPARATORS, FALSE);
					break;
			}
			/* chop off linefeed from end of input line */
			if (*(inlinnow+inlen) == '\n')
				*(inlinnow+inlen) = '\0';
			if (postype != REGEXPR)
				valscn[0] = strdup(inlinnow);
			if (isdebug) {
				logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "input line is: @%s@", inlinnow);
				dumpvalnows(valnow);
			}
			/* build keystr for this line */
			if (KEYS && !HSTERR) {
				vlp = valnow;
				concat = FALSE;
				foreach (kp, KEYS) {
					switch (VARTYP(kp->vrndx)) {
						case SCALAR:
							if (!concat) {
								strcpy(keystr,vlp->VALSCL(kp->vrndx,0));
								concat = TRUE;
							} else {
								strcat(keystr,":");
								strcat(keystr,vlp->VALSCL(kp->vrndx,0));
							}
							break;
						case ARRAY:
							if (!concat) {
								strcpy(keystr,vlp->VALARY(kp->vrndx,kp->arndx));
								concat = TRUE;
							} else {
								strcat(keystr,":");
								strcat(keystr,vlp->VALARY(kp->vrndx,kp->arndx));
							}
							break;
						case NIL:
							break;
						default:
							logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, unknown vartype for var %d", ALMNAM, kp->vrndx);
							STATUS = INACTIVE;
							goto _QUIT;
							break;
					}
				}
				/* lookup this keystr in the alert history */
				valhst = lookupval(keystr, VALHST);
			} else if ((HSTTYP & (FREEVAL|BEGINVAL|ENDVAL)) && !HSTERR)
				valhst = valglb;
			else
				valhst = (val *) NULL;
			_REDO:
			rulcurr = alarmcurr->rules;
			while (rulcurr) {
				_AGAIN:
				if (isdebug)
					logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "RULE");
				stmcurr = rulcurr->statements;
				while (stmcurr) {
					switch (dostm(stmcurr->type)) {
						case TRUE:
							stmcurr = stmcurr->tnext;
							break;
						case FALSE:
							stmcurr = stmcurr->fnext;
							break;
						case AGAIN:	/* redo current rule */
							goto _AGAIN;
							break;
						case LEAVE:	/* leave current rule */
							goto _LEAVE;
							break;
						case REDO:	/* redo current line */
							goto _REDO;
							break;
						case NEXT:	/* go on to next line */
							goto _NEXT;
							break;
						case SKIP:	/* go ahead n lines */
							goto _SKIP;
							break;
						case LAST:	/* break to end */
							if (inlinprc) {
								inlintmp = inlinprc;
								inlinprc = inlinnow;
								inlinnow = inlintmp;
							} else {
								inlinprc = inlin[0];
								inlinnow = inlin[1];
							}
							goto _LAST;
							break;
						case QUIT:	/* go on to next alarm */
						default:
							goto _QUIT;
							break;
					}
				}
				_LEAVE:
				rulcurr = rulcurr->next;
			}
			_NEXT:
			_SKIP:
			/* write valnows to hst file */
			if (KEYS /* && !rulcurr */) {
				vlp = valnow;
				for (i=1; i<=MAXVARS; i++) {
					if (NEWHST(i) && (VALTYP(i) & (FIXEDVAL|FREEVAL))) {
						switch (VARTYP(i)) {
							case SCALAR:
								if (vlp->VALSCL(i,0))
									fprintf(hstout, "V %d %d %s\n", i, 0, vlp->VALSCL(i,0));
								break;
							case ARRAY:
								for (j=1; j<=MAXARNDX(i); j++) {
									if (vlp->SUBARY(i,j) && vlp->VALARY(i,j))
										fprintf(hstout, "V %d %d %s\n", i, j, vlp->VALARY(i,j));
								}
								break;
							case NIL:
								break;
							default:
								logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, unknown vartype for var %d", ALMNAM, i);
								STATUS = INACTIVE;
								goto _QUIT;
								break;
						}
					}
				}
			}
			if (valhst && KEYS && (valhst != valglb))
				freevalhst(valhst);
			isprc = TRUE;	/* successfully processed first line
					   of input, now have preceding val
					 */
			if (inlinprc) {
				inlintmp = inlinprc;
				inlinprc = inlinnow;
				inlinnow = inlintmp;
			} else {
				inlinprc = inlin[0];
				inlinnow = inlin[1];
			}
		} /* while (fgets(inlinnow, MAXLINE, infile) != NULL) */
	}
	_LAST:
	eoi = TRUE;
	inlinnow = (char *) NULL;
	inlen = 0;
	valtmp = valprc;
	valprc = valnow;
	valnow = valtmp;
	valhst = valglb;
	freevals(&valnow, FIXEDVAL|FREEVAL, FALSE);
	/* carry forward free vals from last input line */
	for (i=1; i<=MAXVARS; i++) {
		if (VALTYP(i) & FREEVAL) {
			switch (VARTYP(i)) {
				case SCALAR:
					if (valprc->VALSCL(i,0))
						valnow->VALSCL(i,0) = strdup(valprc->VALSCL(i,0));
					break;
				case ARRAY:
					for (j=1; j<=MAXARNDX(i); j++) {
						if (valprc->SUBARY(i,j) && valprc->VALARY(i,j)) {
							if (! valnow->SUBARY(i,j))
								initchrsubary(&valnow->SUBARY(i,j));
							valnow->VALARY(i,j) = strdup(valprc->VALARY(i,j));
						}
					}
					break;
				case NIL:
					break;
				default:
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, unknown vartype for var %d", ALMNAM, i);
					STATUS = INACTIVE;
					goto _QUIT;
					break;
			}
		}
	}
	if (alarmcurr->end) {
		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "END");
		stmcurr = alarmcurr->end->statements;
		while (stmcurr) {
			switch (dostm(stmcurr->type)) {
				case TRUE:
					stmcurr = stmcurr->tnext;
					break;
				case FALSE:
					stmcurr = stmcurr->fnext;
					break;
				case AGAIN:
					stmcurr = alarmcurr->end->statements;
					break;
				case QUIT:
					goto _QUIT;
					break;
				default:
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, invalid dostm() return at end", ALMNAM);
					STATUS = INACTIVE;
					goto _QUIT;
					break;
			}
		}
	}
	if (infile && (innum == 0) && alarmcurr->rules)
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in scan(), %s, no input data", ALMNAM);
	_QUIT:
	/* close any open files */
	for (i=1; i<=MAXVARS; i++) {
		if (((DATTYP(i) == FILEHANDLE) || (DATTYP(i) == PROCHANDLE)) && FLHNDL(i)) {
			if (fclose(FLHNDL(i)))
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scan(), %s, fclose() error for file handle %s", ALMNAM, VARNAM(i));
		}
	}
	if (infile)
		scanclose(infile, pid);
	/* write globals to history file */
	vlp = valnow;
	for (i=1; i<=MAXVARS; i++) {
		if (NEWHST(i) && (VALTYP(i) & (BEGINVAL|FREEVAL|ENDVAL))) {
			switch (VARTYP(i)) {
				case SCALAR:
					if (vlp->VALSCL(i,0))
						fprintf(hstout, "G %d %d %s\n", i, 0, vlp->VALSCL(i,0));
					break;
				case ARRAY:
					for (j=1; j<=MAXARNDX(i); j++) {
						if (vlp->SUBARY(i,j) && vlp->VALARY(i,j))
							fprintf(hstout, "G %d %d %s\n", i, j, vlp->VALARY(i,j));
					}
					break;
				case NIL:
					break;
				default:
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scan(), %s, unknown vartype for var %d", ALMNAM, i);
					STATUS = INACTIVE;
					break;
			}
		}
	}
	switch (inputtype) {
		case NOINPUT:
			break;
		case PROCINPUT:
			break;
		case LOGINPUT:
			if (inputsource)
				fprintf(hstout, "L %ld %ld %ld %s\n", (unsigned long)inocurr, (unsigned long)sizecurr, posend, inputsource);
			break;
		case FILEINPUT:
			break;
	}
	return 0;
}

/* scan the current input line, or the string passed in the #parse()
   function; break down the line into component parts, with white space
   as the default separator, or using the specified separator char(s);
   assign scanned vals to the appropriate var index;
   scanfields() is for dat statements of the form "dat #foo 1",
   or #parse() invocations of the form "#parse($rdlin)"
 */
int
scanfields (char *line, char *separators, int isparsefn)
{
  	static char *buf;
	unsigned char s;
	register int i, j;
	int col;
	int f, fn;
	register char c;
	int fields;
	int done;
	int nspsep;

	nspsep = FALSE;
	fields = 0;
	s = OUTDAT;
	i = 0;
	col = 0;
	f = 1;
	buf = strdup(line);
	while ((c = *(line+col))) {
		if ((isspace((int)c) && issep(c, separators)) || (c == '\n')) {
			switch (s) {
				case INDAT:
				case BEGINDAT:
					s = ENDDAT;
					break;
				default:
					s = OUTDAT;
					break;
			}
		} else if (issep(c, separators)) {
			nspsep = TRUE;
			/* in all cases, we set s to ENDDAT! */
			switch (s) {
				/*
				case INDAT:
					s = ENDDAT;
					break;
				case BEGINDAT:
					s = ENDDAT;
					break;
				case ENDDAT:
					s = ENDDAT;
					break;
				case OUTDAT:
					s = ENDDAT;
					break;
				*/
				default:
					s = ENDDAT;
					break;
			}
		} else {	/* is not eol, space, tab, or sep */
			switch (s) {
				case OUTDAT:
				case ENDDAT:
					s = BEGINDAT;
					break;
				default:
					s = INDAT;
					break;
			}
		}
		switch (s) {	
			case BEGINDAT:
			case INDAT:
				*(buf+i++) = c;
				break;
			case ENDDAT:
				*(buf+i) = '\0';
				FREE(valscn[f]);
				valscn[f] = strdup(buf);
				i = 0;
				f++;
				break;
			case OUTDAT:
				if (c == '\n') {
					*(buf+i) = '\0';
				}
				break;
		}
		col++;
	}
	if ((s == ENDDAT) || !strcmp(line, "\n"))
		f--;
	if (!valscn[f]) {
		/* if we are parsing $command() output, which doesn't end in a
		   linefeed, we have to terminate the last buf string here
		 */
		if (isparsefn)
			*(buf+i) = '\0';
		if (isparsefn || nspsep)
			valscn[f] = strdup(buf);
		else
			f--;
	}
	if ((DATCNT == 0) || isparsefn)	{ /* no dat statements */
		free(buf);
		return f;
	}
	if (f < MAXDATNDX) {	/* f is # of dat valscn parsed */
		free(buf);
		return ERROR;
	}
	done = FALSE;
	for (i=1; i<=MAXVARS; i++) {
		if (done) break;
		if (VALTYP(i) & FIXEDVAL) {
			switch (VARTYP(i)) {
				case SCALAR:
					fn = FLDNUM(i,0);
					if (fn == 0)
						continue;
					else if (fn < 0)
						fn = f+fn+1;
					if (TRIM(i,0) == BOTH)
						valnow->VALSCL(i,0) = trim(valscn[fn],BOTH);
					else
						valnow->VALSCL(i,0) = strdup(valscn[fn]);
					if (++fields >= DATCNT)
						done = TRUE;
					break;
				case ARRAY:
					for (j=1; j<=MAXARNDX(i); j++) {
						if (done) break;
						fn = FLDNUM(i,j);
						if (fn == 0)
							continue;
						else if (fn < 0)
							fn = f+fn+1;
						if (! valnow->SUBARY(i,j))
							initchrsubary(&valnow->SUBARY(i,j));
						if (TRIM(i,j) == BOTH)
							valnow->VALARY(i,j) = trim(valscn[fn],BOTH);
						else
							valnow->VALARY(i,j) = strdup(valscn[fn]);
						if (++fields >= DATCNT)
							done = TRUE;
					}
					break;
				case NIL:
					break;
				default:
					logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scanfields(), %s, unknown vartype for var %d", ALMNAM, i);
					STATUS = INACTIVE;
					free(buf);
					return 0;
					break;
			}
		}
	}
	free(buf);
	return fields;
}

/* scan the current input line; break down the line into component parts,
   according to their column position; assign scanned vals to the appropriate
   var index;
   scancolumns() is for dat statements of the form "dat #foo 10,15"
 */
int
scancolumns (char *line)
{
	static char *buf;
	unsigned char s;
	register int i;
	int col;
	int f;
	register char c;
	ky *scvar;
	int begincol, endcol;

	i = 0;
	col = 0;
	f = 1;
	buf = strdup(line);
	scvar = SCVARS;
	switch (VARTYP(scvar->vrndx)) {
		case SCALAR:
			begincol = BEGINCOL(scvar->vrndx,0);
			endcol = ENDCOL(scvar->vrndx,0);
			break;
		case ARRAY:
			begincol = BEGINCOL(scvar->vrndx,scvar->arndx);
			endcol = ENDCOL(scvar->vrndx,scvar->arndx);
			break;
		case NIL:
		default:
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scancolumns(), %s, unknown or bad vartype for var %d", ALMNAM, scvar->vrndx);
			STATUS = INACTIVE;
			free(buf);
			return 0;
			break;
	}
	if (!VARNAM(f)) {
		free(buf);
		return 0;
	}
	while ((c = *(line+col))) {
		col++;
		if (c == '\n')
			s = ENDDAT;
		else if (col < begincol)
			s = OUTDAT;
		else if (col == begincol)
			s = BEGINDAT;
		else if (col <= endcol)
			s = INDAT;
		else if (col == endcol+1)
			s = ENDDAT;
		else
			s = OUTDAT;
		switch (s) {
			case BEGINDAT:
			case INDAT:
				*(buf+i++) = c;
				break;
			case ENDDAT:
				*(buf+i) = '\0';
				valscn[f] = strdup(buf);
				switch (VARTYP(scvar->vrndx)) {
					case SCALAR:
						if (TRIM(scvar->vrndx,0) == BOTH)
							valnow->VALSCL(scvar->vrndx,0) = trim(valscn[f],BOTH);
						else
							valnow->VALSCL(scvar->vrndx,0) = strdup(valscn[f]);
						break;
					case ARRAY:
						if (! valnow->SUBARY(scvar->vrndx,scvar->arndx))
							initchrsubary(&valnow->SUBARY(scvar->vrndx,scvar->arndx));
						if (TRIM(scvar->vrndx,scvar->arndx) == BOTH)
							valnow->VALARY(scvar->vrndx,scvar->arndx) = trim(valscn[f],BOTH);
						else
							valnow->VALARY(scvar->vrndx,scvar->arndx) = strdup(valscn[f]);
						break;
					case NIL:
					default:
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scancolumns(), %s, unknown or bad vartype for var %d", ALMNAM, scvar->vrndx);
						STATUS = INACTIVE;
						free(buf);
						return 0;
						break;
				}
				i = 0;
				if (++f > DATCNT) {
					free(buf);
					return --f;
				}
				scvar = scvar->next;
				if (!scvar) {
					free(buf);
					return --f;
				}
				switch (VARTYP(scvar->vrndx)) {
					case SCALAR:
						begincol = BEGINCOL(scvar->vrndx,0);
						endcol = ENDCOL(scvar->vrndx,0);
						break;
					case ARRAY:
						begincol = BEGINCOL(scvar->vrndx,scvar->arndx);
						endcol = ENDCOL(scvar->vrndx,scvar->arndx);
						break;
					case NIL:
					default:
						logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in scancolumns(), %s, unknown or bad vartype for var %d", ALMNAM, scvar->vrndx);
						STATUS = INACTIVE;
						free(buf);
						return 0;
						break;
				}
				if (c != '\n')
					col--;
				break;
			case OUTDAT:
				break;
		}
	}
	free(buf);
	return f;
}

/* is the given char in the separators list (defaults to space (" "))?
 */
int
issep (int c, char *seps)
{
	register int i;
	register char sc;

	if (!seps)	/* should never be true */
		return FALSE;
	else {
		i = 0;
		while ((sc = *(seps+i))) {
			if (sc == c)
				return sc;
			i++;
		}
	}
	return FALSE;
}

/* spawn a child process; the child opens the specified input, whether a
   process accessing some file or some other process; the process output,
   both stdout and stderr, is fed back to the parent; doing it this way,
   and by not using a simple popen(), we can capture output to both
   stdout and stderr, giving us useful debugging info that might otherwise
   be lost; on some machines, we coordinate the parent and child via IPC;
   on other machines, we just have the parent sleep() an arbitary amount
   while waiting for the child
 */
FILE *
scanopen (char *cmdstr, pid_t *pid)
{
	int fd[2];
	FILE *fp;
	FILE *logfile;
	int logfd;

	fp = NULL;
	if (pipe(fd) < 0)
		logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in %s, scanopen(), pipe() error", ALMNAM);
	#if TIME_TO_SLEEP == 0
	/* open the ipc channels */
	ipcopen(0);
	#endif
	switch (*pid = fork()) {
		case -1:	/* error */
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scanopen(), %s, fork error", ALMNAM);
			break;
		/* pipe from child to parent */
		case 0:		/* child */
			close(fd[0]);
			logfile = fopen(piktlog, "a");
			logfd = fileno(logfile);
			dup2(logfd, STDERR_FILENO);
			if (fd[1] != STDOUT_FILENO) {
				dup2(fd[1], STDOUT_FILENO);
				close(fd[1]);
			}
			#if TIME_TO_SLEEP == 0
			wait_parent(0);
			tell_parent(0);
			#endif
			execl(SHELL, "sh", "-c", cmdstr, (char *) NULL);
			fclose(logfile);
			_exit(127);
			break;
		default:	/* parent */
			close(fd[1]);
			#if TIME_TO_SLEEP == 0
			tell_child(0);
			wait_child(0);
			#else
			sleep(TIME_TO_SLEEP);	/* this is a kludge for sure */
			#endif
			if ((fp = fdopen(fd[0], "r")) == NULL)
				return NULL;
			else
				return fp;
			break;
	}
	return fp;
}

/* close the scan input stream
 */
void
scanclose (FILE *fp, pid_t pid)
{
	int status;

	if (fclose(fp))
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in scanclose(), %s, fclose() error", ALMNAM);
	waitpid(pid, &status, 0);
	#if TIME_TO_SLEEP == 0
	ipcclose(0);
	#endif
	return;
}

/* strip backslash escape chars from the given scan command; this is to
   ensure consistency of treatment with escape processing done in begin,
   end, and rules sections, where backslash escapes are dealt with by
   the decompose() function
 */
char *
stripesc (char *input)
{
	static char output[MAXLINE+1];
	char c;
	register int i, j;
	register int escape;

	escape = FALSE;
	i = j = 0;
	while ((c = *(input+i))) {
		i++;
		if ((c == '\\') && !escape) {
			escape = TRUE;
			continue;
		}
		output[j++] = c;
		escape = FALSE;
	}
	output[j] = '\0';
	return output;
}

/* for input str, replace all double quotes (") with DQTPROXY,
   and all single quotes (') with SQTPROXY, returning strlen
 */
int
dequote (char *str)
{
	register int i;
	register char c;

	i = 0;
	while ((c = *(str+i))) {
		if (c == '\"')
			*(str+i) = DQTPROXY;
		else if (c == '\'')
			*(str+i) = SQTPROXY;
		i++;
	}
	return i;
}

/* for input str, replace all DQTPROXY with double quotes ("),
   and all SQTPROXY with single quotes ('), returning strlen
 */
int
requote (char *str)
{
	register int i;
	register char c;

	if (!str)
		return ERROR;
	i = 0;
	while ((c = *(str+i))) {
		if (c == DQTPROXY)
			*(str+i) = '\"';
		else if (c == SQTPROXY)
			*(str+i) = '\'';
		i++;
	}
	return i;
}

/*  garbage collect valscn array
 */
void
freevalscn (void)
{
	register int i;

	for (i=0; i<=matches; i++) {
		if (valscn[i]) {
			free(valscn[i]);
			valscn[i] = (char *) NULL;
		}
	}
}
